# -*- coding: utf-8 -*-
"""
This file contains a settings dialog for the qudi main GUI.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtCore, QtWidgets


class SettingsDialog(QtWidgets.QDialog):
    """
    Custom QDialog widget for configuration of the qudi main GUI.
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.setWindowTitle('Qudi: Main GUI settings')

        # Create main layout
        # Add widgets to layout and set as main layout
        layout = QtWidgets.QGridLayout()
        layout.setRowStretch(1, 1)
        self.setLayout(layout)

        # Create widgets and add them to the layout
        self.font_size_spinbox = QtWidgets.QSpinBox()
        self.font_size_spinbox.setObjectName('fontSizeSpinBox')
        self.font_size_spinbox.setMinimum(5)
        self.font_size_spinbox.setValue(10)
        label = QtWidgets.QLabel('Console font size:')
        label.setObjectName('fontSizeLabel')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.font_size_spinbox, 0, 1)

        self.show_error_popups_checkbox = QtWidgets.QCheckBox()
        self.show_error_popups_checkbox.setObjectName('showErrorPopupsCheckbox')
        self.show_error_popups_checkbox.setChecked(True)
        label = QtWidgets.QLabel('Show error popups:')
        label.setObjectName('showErrorPopupsLabel')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        layout.addWidget(label, 1, 0)
        layout.addWidget(self.show_error_popups_checkbox, 1, 1)

        buttonbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok
                                               | QtWidgets.QDialogButtonBox.Cancel
                                               | QtWidgets.QDialogButtonBox.Apply)
        buttonbox.setOrientation(QtCore.Qt.Horizontal)
        layout.addWidget(buttonbox, 2, 0, 1, 2)

        # Add internal signals
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)
        buttonbox.button(buttonbox.Apply).clicked.connect(self.accepted)



# -*- coding: utf-8 -*-
"""
Descriptor objects that can be used to simplify common tasks related to object attributes.

Copyright (c) 2023, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['BaseAttribute', 'DefaultAttribute', 'ReadOnlyAttribute', 'TypedAttribute',
           'CheckedAttribute', 'DefaultMixin', 'ReadOnlyMixin', 'TypedMixin', 'ValidateMixin']

from typing import Any, Optional, Iterable, Type, Callable, Union
from inspect import isclass, isfunction


class DefaultMixin:
    """Mixin for BaseAttribute introducing optional default value behaviour in __get__.
    If no default value is specified, fall back to raising AttributeError.
    """
    _no_default = object()  # unique placeholder

    def __init__(self, default: Optional[Any] = _no_default, **kwargs):
        super().__init__(**kwargs)
        self.default = default

    def __get__(self, instance, owner):
        try:
            return super().__get__(instance, owner)
        except AttributeError:
            if self.default is self._no_default:
                raise
            return self.default

    def __delete__(self, instance):
        try:
            super().__delete__(instance)
        except AttributeError:
            pass


class ReadOnlyMixin:
    """Mixin for BaseAttribute introducing read-only access."""
    def __delete__(self, instance):
        raise AttributeError('Read-only attribute can not be deleted')

    def __set__(self, instance, value):
        raise AttributeError('Read-only attribute can not be overwritten')

    def set_value(self, instance: object, value: Any) -> None:
        super().__set__(instance, value)


class TypedMixin:
    """Mixin for BaseAttribute introducing optional type checking via isinstance builtin."""
    def __init__(self, valid_types: Optional[Iterable[Type]] = None, **kwargs):
        super().__init__(**kwargs)
        self.valid_types = None if valid_types is None else tuple(valid_types)
        if self.valid_types and not all(isclass(typ) for typ in self.valid_types):
            raise TypeError('valid_types must be iterable of types (classes)')

    def __set__(self, instance, value):
        self.check_type(value)
        super().__set__(instance, value)

    def check_type(self, value: Any) -> None:
        if self.valid_types and not isinstance(value, self.valid_types):
            raise TypeError(
                f'Value must be of type(s) [{", ".join(t.__name__ for t in self.valid_types)}]'
            )


class ValidateMixin:
    """Mixin for BaseAttribute introducing optional validation via registering static and/or
    bound validator methods.
    Bound methods are best registered via the "validator" decorator (cooperative with
    staticmethod/classmethod decorator).
    """
    def __init__(self,
                 static_validators: Optional[Iterable[Callable[[Any], None]]] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.static_validators = list() if static_validators is None else list(static_validators)
        self.bound_validators = list()
        if not all(callable(val) for val in self.static_validators):
            raise TypeError('static_validators must be iterable of callables')

    def __set__(self, instance, value):
        self.validate(value, instance)
        super().__set__(instance, value)

    def validator(self,
                  func: Union[staticmethod, classmethod, Callable[[Any], None]]
                  ) -> Union[staticmethod, classmethod, Callable[[Any], None]]:
        """Decorator to register either a static or bound validator."""
        # Use function reference directly if static
        if isinstance(func, staticmethod):
            self.static_validators.append(func.__func__)
            return func

        # In case of bound methods (class/instance) just use the attribute name string
        if isinstance(func, classmethod):
            func_obj = func.__func__
        elif isfunction(func):
            func_obj = func
            if func_obj.__qualname__ == func_obj.__name__:
                # Not a class member, thus probably static
                self.static_validators.append(func)
                return func
        else:
            raise TypeError('validator must either be function, staticmethod or classmethod object')

        # Take care of name mangling for private members
        if func_obj.__name__.startswith('__'):
            cls_name = func_obj.__qualname__.rsplit('.', 1)[0]
            self.bound_validators.append(f'_{cls_name}{func_obj.__name__}')
        else:
            self.bound_validators.append(func_obj.__name__)
        return func

    def validate(self, value: Any, instance: Optional[Any] = None) -> None:
        try:
            for func in self.static_validators:
                func(value)
            for func_name in self.bound_validators:
                try:
                    func = getattr(instance, func_name)
                except AttributeError:
                    raise AttributeError(
                        f'Registered bound validator "{func_name}" not found in {instance}'
                    ) from None
                func(value)
        except Exception as err:
            raise ValueError(f'Value "{value}" did not pass validation') from err


class BaseAttribute:
    """Base descriptor class implementing trivial get/set/delete behaviour for an instance
    attribute.
    """
    def __init__(self):
        super().__init__()
        self.attr_name = None

    def __set_name__(self, owner, name):
        self.attr_name = name

    def __get__(self, instance, owner):
        try:
            return instance.__dict__[self.attr_name]
        except KeyError:
            raise AttributeError(self.attr_name) from None
        except AttributeError:
            return self

    def __delete__(self, instance):
        try:
            del instance.__dict__[self.attr_name]
        except KeyError:
            raise AttributeError(self.attr_name) from None

    def __set__(self, instance, value):
        instance.__dict__[self.attr_name] = value


class DefaultAttribute(DefaultMixin, BaseAttribute):
    """Attribute that can be given a default value which is used if not explicitly initialized by
    the instance.

    Example usage:

        class Test:
            variable_a = DefaultAttribute(42)
            variable_b = DefaultAttribute()
            def __init__(self):
                self.variable_b = self.variable_a - 42
                assert self.variable_a == 42
                assert self.variable_b == 0
    """
    def __init__(self, default: Optional[Any] = DefaultMixin._no_default):
        super().__init__(default=default)


class ReadOnlyAttribute(ReadOnlyMixin, DefaultAttribute):
    """Extension of DefaultAttribute to be read-only. A non-default value can be set by calling
    "set_value(instance, value)" on the descriptor instance.

    Example usage:

        class Test:
            variable_a = ReadOnlyAttribute(42)
            variable_b = ReadOnlyAttribute()
            def __init__(self):
                self.__class__.variable_b.set_value(self, self.variable_a - 42)
                assert self.variable_a == 42
                assert self.variable_b == 0
                # The following would raise an AttributeError
                # self.variable_b = 0
    """
    pass


class TypedAttribute(TypedMixin, DefaultAttribute):
    """Extension of DefaultAttribute including type checking via isinstance. A given default
    value is not type-checked.

    Example usage:

        class Test:
            variable_a = TypedAttribute([int, float])
            variable_b = TypedAttribute([str], None)
            def __init__(self):
                assert self.variable_b is None
                self.variable_a = 42
                self.variable_b = 'hello world'
                assert self.variable_a == 42
                assert self.variable_b == 'hello world'
                # The following would raise TypeError
                # self.variable_a = self.variable_b = None
    """
    def __init__(self,
                 valid_types: Optional[Iterable[Type]] = None,
                 default: Optional[Any] = DefaultAttribute._no_default):
        super().__init__(valid_types=valid_types, default=default)


class CheckedAttribute(TypedMixin, ValidateMixin, DefaultAttribute):
    """Extension of DefaultAttribute including optional validation via static or bound validator
    methods as well as optional type checking via "isinstance".
    A given default value is not validated. Type checking is performed before validation.
    Register bound validator methods via the CheckedAttribute.validator decorator. This decorator
    can be combined with classmethod/staticmethod decorators in any order.

    Example usage:

        def my_static_validator(value):
            if not (0 <= value <= 100):
                raise ValueError('Value must be number between 0 and 100')

        class Test:
            variable_a = CheckedAttribute([my_static_validator], [int, float], 0)
            variable_b = CheckedAttribute(valid_types=[str])
            _valid_strings = ['A', 'B', 'C']

            def __init__(self):
                self.variable_a = 66.7
                self.variable_b = 'B'
                assert self.variable_a == 66.7
                assert self.variable_b == 'B'
                # The following would raise ValueError
                # self.variable_a = 101
                # self.variable_b = 'D'

            @variable_b.validator
            @classmethod
            def _validate_variable_b(cls, value):
                if value not in cls._valid_strings:
                    raise ValueError(f'Invalid string. Valid strings are: {cls._valid_strings}')
    """
    def __init__(self,
                 static_validators: Optional[Iterable[Callable[[Any], None]]] = None,
                 valid_types: Optional[Iterable[Type]] = None,
                 default: Optional[Any] = DefaultAttribute._no_default):
        super().__init__(static_validators=static_validators,
                         valid_types=valid_types,
                         default=default)



# -*- coding: utf-8 -*-

"""
This file contains base and meta class for data fit model classes for qudi based on the lmfit
package. Also contains an estimator decorator for fit models to name estimator methods.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('estimator', 'FitCompositeModelBase', 'FitCompositeModelMeta', 'FitModelBase',
           'FitModelMeta')

import inspect
from abc import ABCMeta, abstractmethod
from lmfit import Model, CompositeModel


def estimator(name):
    assert isinstance(name, str) and name, 'estimator name must be non-empty str'

    def _decorator(func):
        assert callable(func), 'estimator must be callable'
        params = tuple(inspect.signature(func).parameters)
        assert len(params) == 3, \
            'estimator must be bound method with 2 positional parameters. First parameter is the ' \
            'y data array to use and second parameter is the corresponding independent variable.'
        func._estimator_name = name
        func._estimator_independent_var = params[2]
        return func

    return _decorator


class FitModelMeta(ABCMeta):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)

        # collect marked estimator methods in a dict "_estimators" and attach it to created class.
        # NOTE: Estimators defined in base classes will not be taken into account. If you want to
        # do inheritance shenanigans with these fit model classes, you need to manually handle this
        # in the implementation.
        # Generally one can not assume parent estimators to be valid for a subclass.
        cls._estimators = {attr._estimator_name: attr for attr in attrs.values() if
                           hasattr(attr, '_estimator_name')}
        independent_vars = {e._estimator_independent_var for e in cls._estimators.values()}
        assert len(independent_vars) < 2, \
            'More than one independent variable name encountered in estimators. Use only the ' \
            'independent variable name that has been used in the Models "_model_function".'


class FitCompositeModelMeta(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)

        # collect marked estimator methods in a dict "_estimators" and attach it to created class.
        # NOTE: Estimators defined in base classes will not be taken into account. If you want to
        # do inheritance shenanigans with these fit model classes, you need to manually handle this
        # in the implementation.
        # Generally one can not assume parent estimators to be valid for a subclass.
        cls._estimators = {attr._estimator_name: attr for attr in attrs.values() if
                           hasattr(attr, '_estimator_name')}
        independent_vars = {e._estimator_independent_var for e in cls._estimators.values()}
        assert len(independent_vars) < 2, \
            'More than one independent variable name encountered in estimators. Use only the ' \
            'independent variable name that has been used in the Models "_model_function".'


class FitModelBase(Model, metaclass=FitModelMeta):
    """ ToDo: Document
    """

    def __init__(self, **kwargs):
        kwargs['name'] = self.__class__.__name__
        super().__init__(self._model_function, **kwargs)
        assert len(self.independent_vars) == 1, \
            'Qudi fit models must contain exactly 1 independent variable.'
        # Shadow FitModelBase._estimators with a similar dict containing the bound method objects.
        # This instance-level dict has read-only access via property "estimators"
        self._estimators = {name: getattr(self, e.__name__) for name, e in self._estimators.items()}

    @property
    def estimators(self):
        """ Read-only dict property holding available estimator names as keys and the corresponding
        estimator methods as values.

        Returns
        -------
        dict
            Available estimator methods (values) with corresponding names (keys).

        """
        return self._estimators.copy()

    @staticmethod
    @abstractmethod
    def _model_function(x, **kwargs):
        """ ToDo: Document
        """
        raise NotImplementedError('FitModel object must implement staticmethod "_model_function".')


class FitCompositeModelBase(CompositeModel, metaclass=FitCompositeModelMeta):
    """ ToDo: Document
    """

    def __init__(self, *args, **kwargs):
        kwargs['name'] = self.__class__.__name__
        super().__init__(*args, **kwargs)
        assert len(self.independent_vars) == 1, \
            'Qudi fit models must contain exactly 1 independent variable.'
        # Shadow FitCompositeModelBase._estimators with a similar dict containing the bound method
        # objects. This instance-level dict has read-only access via property "estimators"
        self._estimators = {name: getattr(self, e.__name__) for name, e in self._estimators.items()}

    @property
    def estimators(self):
        """ Read-only dict property holding available estimator names as keys and the corresponding
        estimator methods as values.

        Returns
        -------
        dict
            Dictionary where keys are estimator names and values are corresponding estimator methods.

        """
        return self._estimators.copy()



# -*- coding: utf-8 -*-
"""
This file contains the Qudi logging handler objects.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('LogSignalHandler', 'LogTableModelHandler', 'qt_message_handler')

import logging
from PySide2 import QtCore

from .records_model import LogRecordsTableModel


class QtSignaller(QtCore.QObject):
    """Just a bare Qt QObject containing a signal.
    """
    sigSignal = QtCore.Signal(object)


class LogSignalHandler(logging.Handler):
    """Logging handler that emits a Qt signal when a log entry is registered.
    """
    def __init__(self, level=logging.NOTSET):
        super().__init__(level=level)
        self.__qt_signaller = QtSignaller()

    @property
    def sigRecordLogged(self):
        return self.__qt_signaller.sigSignal

    def emit(self, record):
        """Emit a signal when logging.Handler emits a new log record.
        """
        self.__qt_signaller.sigSignal.emit(record)


class LogTableModelHandler(logging.Handler):
    """Logging handler that stores each log record in a QAbstractTableModel.
    """
    def __init__(self, level=logging.INFO, max_records=10000):
        if level < logging.DEBUG:
            level = logging.DEBUG
        super().__init__(level=level)
        self.__qt_signaller = QtSignaller()
        self.table_model = LogRecordsTableModel(max_records=max_records)
        self.__qt_signaller.sigSignal.connect(self.table_model.add_record,
                                              QtCore.Qt.QueuedConnection)

    def emit(self, record):
        """Store the log record information in the table model.
        """
        self.__qt_signaller.sigSignal.emit(record)


def qt_message_handler(msg_type, context, msg):
    """
    A message handler handling Qt5 messages.
    """
    logger = logging.getLogger('Qt')
    if msg_type == QtCore.QtDebugMsg:
        logger.debug(msg)
    elif msg_type == QtCore.QtInfoMsg:
        logger.info(msg)
    elif msg_type == QtCore.QtWarningMsg:
        logger.warning(msg)
    elif msg_type == QtCore.QtCriticalMsg:
        logger.critical(msg)
    else:
        import traceback
        traceback_str = ''.join(traceback.format_stack())
        logger.critical(f'Fatal error occurred: {msg}\nTraceback:\n{traceback_str}')



# -*- coding: utf-8 -*-

"""
This file contains the qudi application watchdog.
Derived from the ACQ4 project.

Copyright (c) 2010, Luke Campagnola.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['AppWatchdog']

import os
import sys
import signal
from PySide2 import QtCore
from qudi.core.parentpoller import ParentPollerWindows, ParentPollerUnix
from qudi.core.logger import get_logger

logger = get_logger(__name__)


class AppWatchdog(QtCore.QObject):
    """This class periodically runs a function for debugging and handles application exit.
    """

    def __init__(self, quit_function):
        super().__init__()
        # Run python code periodically to allow interactive debuggers to interrupt the qt event loop
        self.__timer = QtCore.QTimer()
        self.__timer.timeout.connect(self.do_nothing)
        self.__timer.start(1000)

        # Listen to SIGINT and terminate
        if sys.platform == 'win32':
            signal.signal(signal.SIGINT, lambda *args: quit_function())

        if 'QUDI_PARENT_PID' not in os.environ:
            self.parent_handle = None
            self.parent_poller = None
            logger.warning('Qudi running unsupervised. Restart will not work. Instead Qudi will '
                           'exit with exitcode 42.')
        else:
            self.parent_handle = int(os.environ['QUDI_PARENT_PID'])
            if sys.platform == 'win32':
                self.parent_poller = ParentPollerWindows(self.parent_handle, quit_function)
            else:
                self.parent_poller = ParentPollerUnix(quit_function)
            self.parent_poller.start()
        return

    @QtCore.Slot()
    def do_nothing(self):
        """This function does nothing for debugging purposes.
        """
        x = 0
        for i in range(100):
            x += i
        return

    def terminate(self) -> None:
        try:
            self.parent_poller.stop()
            self.parent_poller.join()
        except AttributeError:
            pass
        finally:
            self.parent_poller = None
            self.parent_handle = None

        try:
            self.__timer.timeout.disconnect()
            self.__timer.stop()
        finally:
            self.__timer = None




# -*- coding: utf-8 -*-

"""
This file contains a wrapper to display the SpinBox in scientific way

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['DoubleSlider']

from PySide2 import QtCore, QtWidgets


class DoubleSlider(QtWidgets.QSlider):
    """

    """
    doubleValueChanged = QtCore.Signal(float)
    doubleSliderMoved = QtCore.Signal(float)
    doubleRangeChanged = QtCore.Signal(float, float)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._step_number = 1000
        self._minimum_value = 0
        self._maximum_value = 1
        super().setRange(0, self._step_number)
        self.valueChanged.connect(self.__translate_value_changed)
        self.sliderMoved.connect(self.__translate_slider_moved)
        self.rangeChanged.connect(self.__translate_range_changed)
        return

    def setMinimum(self, min_val):
        self._minimum_value = float(min_val)
        return

    def setMaximum(self, max_val):
        self._maximum_value = float(max_val)
        return

    def setRange(self, min_val, max_val):
        self._minimum_value = float(min_val)
        self._maximum_value = float(max_val)
        return

    def minimum(self):
        return self._minimum_value

    def maximum(self):
        return self._maximum_value

    def value(self):
        int_val = super().value()
        return self.minimum() + (self.maximum() - self.minimum()) * (int_val / self._step_number)

    def setValue(self, val):
        max_val = self.maximum()
        min_val = self.minimum()
        val = max(min_val, min(max_val, val))
        int_val = int(round((val - min_val) * self._step_number / (max_val - min_val)))
        super().setValue(int_val)

    @property
    def granularity(self) -> int:
        return self._step_number + 1

    def set_granularity(self, number_of_steps):
        """
        Set the granularity of the slider, i.e. the number of discrete steps within the value range.

        Parameters
        ----------
        number_of_steps : int
            The number of discrete positions the slider has.
        """
        number_of_steps = int(number_of_steps)
        if number_of_steps < 1:
            raise ValueError('Number of steps must be larger than 0.')
        self._step_number = number_of_steps - 1  # Include 0 as position
        super().setRange(0, self._step_number)
        return

    @QtCore.Slot(int)
    def __translate_value_changed(self, int_val):
        self.doubleValueChanged.emit(
            self.minimum() + (self.maximum() - self.minimum()) * (int_val / self._step_number))
        return

    @QtCore.Slot(int)
    def __translate_slider_moved(self, int_val):
        self.doubleSliderMoved.emit(
            self.minimum() + (self.maximum() - self.minimum()) * (int_val / self._step_number))
        return

    @QtCore.Slot()
    def __translate_range_changed(self):
        self.doubleRangeChanged.emit(self.minimum(), self.maximum())
        return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import subprocess
import sys
import os


def main():
    """
    """
    myenv = os.environ.copy()

    # Set parent process PID as environment variable for qudi main process
    if sys.platform == 'win32':
        try:
            from _winapi import DuplicateHandle, GetCurrentProcess
            from _winapi import DUPLICATE_SAME_ACCESS
        except ImportError:
            from _subprocess import DuplicateHandle, GetCurrentProcess
            from _subprocess import DUPLICATE_SAME_ACCESS
        pid = GetCurrentProcess()
        handle = DuplicateHandle(pid, pid, pid, 0, True, DUPLICATE_SAME_ACCESS)
        myenv['QUDI_PARENT_PID'] = str(int(handle))
    else:
        myenv['QUDI_PARENT_PID'] = str(os.getpid())

    argv = [sys.executable, '-m', 'core'] + sys.argv[1:]
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    while True:
        process = subprocess.Popen(argv,
                                   close_fds=False,
                                   env=myenv,
                                   stdin=sys.stdin,
                                   stdout=sys.stdout,
                                   stderr=sys.stderr,
                                   shell=False)
        try:
            retval = process.wait()
            if retval == 0:
                break
            elif retval == 42:
                print('Restarting...')
                continue
            elif retval == 2:
                # invalid commandline argument
                break
            elif retval == -6:
                # called if QFatal occurs
                break
            elif retval == 4:
                print('Import Error: Qudi could not be started due to missing packages.')
                sys.exit(retval)
            else:
                print('Unexpected return value {0}. Exiting.'.format(retval))
                sys.exit(retval)
        except KeyboardInterrupt:
            print('\nQudi terminated by keyboard interrupt event!')
            try:
                process.terminate()
                process.wait()
            except:
                pass
            sys.exit(0)
        except:
            process.kill()
            process.wait()
            raise


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
ToDo: Document

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['FitWidget', 'FitConfigurationWidget', 'FitConfigurationDialog']

import os
import weakref
from PySide2 import QtCore, QtWidgets, QtGui
from qudi.util.datafitting import FitContainer, FitConfigurationsModel, FitConfiguration
from qudi.util.paths import get_artwork_dir
from qudi.util.widgets.scientific_spinbox import ScienDSpinBox
from qudi.util.widgets.separator_lines import HorizontalLine


class FitWidget(QtWidgets.QWidget):
    """
    """

    sigDoFit = QtCore.Signal(str)

    def __init__(self, *args, fit_container=None, **kwargs):
        super().__init__(*args, **kwargs)
        main_layout = QtWidgets.QGridLayout()
        self.setLayout(main_layout)

        self.selection_combobox = QtWidgets.QComboBox()
        self.selection_combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.fit_pushbutton = QtWidgets.QPushButton('Fit')
        self.fit_pushbutton.setMinimumWidth(3 * self.fit_pushbutton.sizeHint().width())
        self.result_label = QtWidgets.QLabel()
        self.result_label.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        self.result_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        main_layout.addWidget(self.fit_pushbutton, 0, 0)
        main_layout.addWidget(self.selection_combobox, 0, 1)
        main_layout.addWidget(self.result_label, 1, 0, 1, 2)
        main_layout.setColumnStretch(1, 1)

        self.fit_pushbutton.clicked.connect(self._fit_clicked)

        self.__fit_container_ref = lambda: None
        if fit_container is not None:
            self.link_fit_container(fit_container)

    def link_fit_container(self, fit_container):
        assert (fit_container is None) or isinstance(fit_container, FitContainer), \
            'Can only link qudi FitContainer instances.'
        old_container = self.__fit_container_ref()
        # disconnect old fit container if present
        if old_container is not None:
            old_container.sigFitConfigurationsChanged.disconnect(self.update_fit_configurations)
            old_container.sigLastFitResultChanged.disconnect(self.update_fit_result)
        # link new fit container
        self.result_label.clear()
        self.selection_combobox.clear()
        self.selection_combobox.addItem('No Fit')
        if fit_container is None:
            self.__fit_container_ref = lambda: None
        else:
            self.__fit_container_ref = weakref.ref(fit_container)
            self.selection_combobox.addItems(fit_container.fit_configuration_names)
            fit_container.sigFitConfigurationsChanged.connect(
                self.update_fit_configurations, QtCore.Qt.QueuedConnection
            )
            fit_container.sigLastFitResultChanged.connect(
                self.update_fit_result, QtCore.Qt.QueuedConnection
            )

    @QtCore.Slot(tuple)
    def update_fit_configurations(self, config_names):
        old_text = self.selection_combobox.currentText()
        self.selection_combobox.clear()
        self.selection_combobox.addItem('No Fit')
        self.selection_combobox.addItems(config_names)
        if old_text in config_names:
            self.selection_combobox.setCurrentText(old_text)

    @QtCore.Slot(str, object)
    def update_fit_result(self, fit_config, fit_result):
        if fit_config:
            if self.selection_combobox.currentText() != fit_config:
                self.selection_combobox.setCurrentText(fit_config)
            container = self.__fit_container_ref()
            if container is not None:
                self.result_label.setText(container.formatted_result(fit_result))
            else:
                self.result_label.setText('')

    @QtCore.Slot()
    def _fit_clicked(self):
        config = self.selection_combobox.currentText()
        self.sigDoFit.emit(config)


class FitConfigurationWidget(QtWidgets.QWidget):
    """
    """
    _sigAddNewConfig = QtCore.Signal(str, str)  # name, model

    def __init__(self, *args, fit_config_model, **kwargs):
        assert isinstance(fit_config_model, FitConfigurationsModel)
        super().__init__(*args, **kwargs)
        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(0, 0, 0, 0)

        icon_dir = os.path.join(get_artwork_dir(), 'icons')

        # Create new fit config editor elements
        self.model_combobox = QtWidgets.QComboBox()
        self.model_combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.model_combobox.addItems(fit_config_model.model_names)
        self.name_lineedit = QtWidgets.QLineEdit()
        self.add_config_toolbutton = QtWidgets.QToolButton()
        self.add_config_toolbutton.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
        self.add_config_toolbutton.setIcon(QtGui.QIcon(os.path.join(icon_dir, 'list-add')))
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(self.model_combobox)
        hlayout.addWidget(self.name_lineedit)
        hlayout.addWidget(self.add_config_toolbutton)
        hlayout.setContentsMargins(0, 0, 0, 0)
        # hlayout.setStretch(0, 1)
        hlayout.setStretch(1, 1)
        main_layout.addLayout(hlayout)

        # Create fit config editor list view
        self.config_listview = FitConfigurationListView()
        self.config_listview.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.config_listview.setModel(fit_config_model)
        main_layout.addWidget(self.config_listview)
        main_layout.setStretch(1, 1)

        self.add_config_toolbutton.clicked.connect(self._add_config_clicked)
        self._sigAddNewConfig.connect(
            fit_config_model.add_configuration, QtCore.Qt.QueuedConnection
        )

    @QtCore.Slot()
    def _add_config_clicked(self):
        model = self.model_combobox.currentText()
        name = self.name_lineedit.text()
        if name and model:
            self.name_lineedit.clear()
            self._sigAddNewConfig.emit(name, model)


class FitConfigurationDialog(QtWidgets.QDialog):
    """
    """

    def __init__(self, *args, fit_config_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('Fit Configuration')
        # create main layout
        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        # create config widget
        self.fit_config_widget = FitConfigurationWidget(fit_config_model=fit_config_model)
        main_layout.addWidget(self.fit_config_widget)
        # create dialog buttonbox
        button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok, QtCore.Qt.Horizontal)
        min_width = QtGui.QFontMetrics(QtGui.QFont()).horizontalAdvance('OK OK OK')
        button_box.buttons()[0].setMinimumWidth(min_width)
        button_box.setCenterButtons(True)
        button_box.accepted.connect(self.accept)
        main_layout.addWidget(button_box)
        current_size = self.sizeHint()
        current_size.setWidth(current_size.width() * 2)
        current_size.setHeight(current_size.height() * 2)
        self.resize(current_size)


class FitConfigurationListView(QtWidgets.QListView):
    """
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setVerticalScrollMode(self.ScrollPerPixel)
        self.setMouseTracking(True)
        self.installEventFilter(self)
        config_item_delegate = _FitConfigurationItemDelegate(parent=self)
        self.setItemDelegate(config_item_delegate)
        self.__previous_index = QtCore.QModelIndex()

    def mouseMoveEvent(self, event):
        curr_index = self.indexAt(event.pos())
        if curr_index != self.__previous_index:
            if curr_index.isValid():
                if self.__previous_index.isValid():
                    self.closePersistentEditor(self.__previous_index)
                self.openPersistentEditor(curr_index)
                self.__previous_index = curr_index
        return super().mouseMoveEvent(event)

    def eventFilter(self, object, event):
        if event.type() == QtCore.QEvent.HoverLeave:
            if not self.geometry().contains(event.pos()):
                if self.__previous_index.isValid():
                    self.closePersistentEditor(self.__previous_index)
                self.__previous_index = QtCore.QModelIndex()
                return True
        return False

    @QtCore.Slot(str)
    def remove_config_clicked(self, config_name):
        if self.__previous_index.isValid():
            self.closePersistentEditor(self.__previous_index)
            self.__previous_index = QtCore.QModelIndex()
        self.model().remove_configuration(config_name)


class _FitConfigPanel(QtWidgets.QWidget):
    """
    """
    sigConfigurationRemovedClicked = QtCore.Signal(str)

    def __init__(self, *args, fit_config, **kwargs):
        assert isinstance(fit_config, FitConfiguration)
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        groupbox = QtWidgets.QGroupBox(fit_config.name)
        font = groupbox.font()
        font.setBold(True)
        font.setPointSize(font.pointSize() + 4)
        groupbox.setFont(font)
        layout.addWidget(groupbox)
        main_layout = QtWidgets.QVBoxLayout()
        groupbox.setLayout(main_layout)

        # add remove button
        icon_dir = os.path.join(get_artwork_dir(), 'icons')
        self._name = fit_config.name
        self.remove_config_toolbutton = QtWidgets.QToolButton()
        self.remove_config_toolbutton.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
        self.remove_config_toolbutton.setIcon(
            QtGui.QIcon(os.path.join(icon_dir, 'list-remove'))
        )
        self.remove_config_toolbutton.clicked.connect(
            lambda: self.sigConfigurationRemovedClicked.emit(self._name)
        )

        # add estimator combobox
        self.estimator_selection_combobox = QtWidgets.QComboBox()
        self.estimator_selection_combobox.addItems(fit_config.available_estimators)
        self.estimator_selection_combobox.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        label = QtWidgets.QLabel('Estimator:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        hlayout = QtWidgets.QHBoxLayout()
        main_layout.addLayout(hlayout)
        hlayout.addWidget(label)
        hlayout.addWidget(self.estimator_selection_combobox)
        hlayout.addStretch(1)
        hlayout.addWidget(self.remove_config_toolbutton)

        # add horizontal line
        main_layout.addWidget(HorizontalLine())

        # add parameters
        param_layout = QtWidgets.QGridLayout()
        main_layout.addLayout(param_layout)
        label = QtWidgets.QLabel('customize?')
        param_layout.addWidget(label, 0, 0)
        label = QtWidgets.QLabel('vary?')
        param_layout.addWidget(label, 0, 2)
        label = QtWidgets.QLabel('init:')
        param_layout.addWidget(label, 0, 3)
        label = QtWidgets.QLabel('min:')
        param_layout.addWidget(label, 0, 4)
        label = QtWidgets.QLabel('max:')
        param_layout.addWidget(label, 0, 5)
        # determine minimum width for SpinBoxes based on font metrics
        min_width = QtGui.QFontMetrics(label.font()).horizontalAdvance('999.999')
        self.parameters_widgets = dict()
        row = 1
        for param_name, param in fit_config.default_parameters.items():
            customize_checkbox = QtWidgets.QCheckBox()
            param_layout.addWidget(customize_checkbox, row, 0)
            label = QtWidgets.QLabel(param_name + ':')
            label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
            param_layout.addWidget(label, row, 1)
            vary_checkbox = QtWidgets.QCheckBox()
            vary_checkbox.setChecked(param.vary)
            customize_checkbox.toggled.connect(vary_checkbox.setEnabled)
            param_layout.addWidget(vary_checkbox, row, 2)
            init_spinbox = ScienDSpinBox()
            init_spinbox.setMinimumWidth(min_width)
            init_spinbox.setRange(param.min, param.max)
            init_spinbox.setValue(param.value)
            customize_checkbox.toggled.connect(init_spinbox.setEnabled)
            param_layout.addWidget(init_spinbox, row, 3)
            min_spinbox = ScienDSpinBox()
            min_spinbox.setMinimumWidth(min_width)
            min_spinbox.setRange(param.min, param.max)
            min_spinbox.setValue(param.min)
            customize_checkbox.toggled.connect(min_spinbox.setEnabled)
            param_layout.addWidget(min_spinbox, row, 4)
            max_spinbox = ScienDSpinBox()
            max_spinbox.setMinimumWidth(min_width)
            max_spinbox.setRange(param.min, param.max)
            max_spinbox.setValue(param.max)
            customize_checkbox.toggled.connect(max_spinbox.setEnabled)
            param_layout.addWidget(max_spinbox, row, 5)
            self.parameters_widgets[param_name] = (
                customize_checkbox, vary_checkbox, init_spinbox, min_spinbox, max_spinbox
            )
            row += 1
        param_layout.setColumnStretch(3, 1)
        param_layout.setColumnStretch(4, 1)
        param_layout.setColumnStretch(5, 1)
        self.update_fit_config(fit_config)

    @property
    def estimator(self):
        return self.estimator_selection_combobox.currentText()

    @property
    def custom_parameters(self):
        parameters = dict()
        for param_name, widgets in self.parameters_widgets.items():
            if widgets[0].isChecked():
                parameters[param_name] = (widgets[1].isChecked(),
                                          widgets[2].value(),
                                          widgets[3].value(),
                                          widgets[4].value())
        return parameters

    def update_fit_config(self, config):
        self.blockSignals(True)
        self.estimator_selection_combobox.setCurrentText(config.estimator)
        custom_params = config.custom_parameters
        for param_name, widgets in self.parameters_widgets.items():
            customize = (custom_params is not None) and (param_name in custom_params)
            widgets[0].setChecked(customize)
            widgets[1].setEnabled(customize)
            widgets[2].setEnabled(customize)
            widgets[3].setEnabled(customize)
            widgets[4].setEnabled(customize)
            if customize:
                param = custom_params[param_name]
                widgets[1].setChecked(param.vary)
                widgets[2].setValue(param.value)
                widgets[3].setValue(param.min)
                widgets[4].setValue(param.max)
        self.blockSignals(False)


class _FitConfigurationItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    """

    def createEditor(self, parent, option, index):
        if index.isValid():
            editor = _FitConfigPanel(parent=parent, fit_config=index.data(QtCore.Qt.DisplayRole))
            editor.setGeometry(option.rect)
            editor.sigConfigurationRemovedClicked.connect(self.parent().remove_config_clicked)
            return editor
        return None

    def setEditorData(self, editor, index):
        if index.isValid():
            editor.update_fit_config(index.data(QtCore.Qt.DisplayRole))

    def setModelData(self, editor, model, index):
        data = (editor.estimator, editor.custom_parameters)
        model.setData(index, data)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)
        return option.rect

    def sizeHint(self, option, index):
        size = _FitConfigPanel(fit_config=index.data(QtCore.Qt.DisplayRole)).sizeHint()
        return size

    def paint(self, painter, option, index):
        painter.save()
        r = option.rect
        painter.translate(r.topLeft())
        widget = _FitConfigPanel(fit_config=index.data(QtCore.Qt.DisplayRole))
        widget.setGeometry(r)
        widget.render(painter, QtCore.QPoint(0, 0), painter.viewport())
        painter.restore()

    def destroyEditor(self, editor, index):
        editor.sigConfigurationRemovedClicked.disconnect()
        self.setModelData(editor, index.model(), index)
        return super().destroyEditor(editor, index)



# -*- coding: utf-8 -*-

"""
QWidget serving as main editor for the global configuration section

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['GlobalEditorWidget']

from PySide2 import QtCore, QtWidgets
from typing import Optional, Mapping, Dict, Union, Any
from qudi.tools.config_editor.global_widgets import GlobalConfigWidget


class GlobalEditorWidget(QtWidgets.QStackedWidget):
    """
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent=parent)

        self.placeholder_label = QtWidgets.QLabel('Please load configuration from file\n'
                                                  'or create a new one.')
        font = self.placeholder_label.font()
        font.setBold(True)
        font.setPointSize(font.pointSize() + 4)
        self.placeholder_label.setFont(font)
        self.placeholder_label.setAlignment(QtCore.Qt.AlignCenter)
        self.placeholder_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                             QtWidgets.QSizePolicy.Expanding)
        self.addWidget(self.placeholder_label)

        self.global_editor_widget = GlobalConfigWidget()
        self.addWidget(self.global_editor_widget)

        self.setCurrentIndex(0)

    @property
    def config(self) -> Union[None, Dict[str, Any]]:
        if self.currentIndex() == 0:
            return None
        else:
            return self.global_editor_widget.config

    def set_config(self, config: Union[None, Dict[str, Any]]) -> None:
        self.global_editor_widget.set_config(config)

    def open_editor(self, config: Union[None, Mapping[str, Any]]) -> None:
        self.global_editor_widget.set_config(config)
        self.setCurrentIndex(1)

    def close_editor(self):
        self.setCurrentIndex(0)



# -*- coding: utf-8 -*-
"""
Jupyter notebook kernel executable file for Qudi.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['install_kernel', 'uninstall_kernel', 'QudiIPythonKernel', 'QudiKernelClient',
           'QudiKernelService']

import os
import sys
import rpyc
import json
import shutil
import logging
import tempfile
from ipykernel.ipkernel import IPythonKernel

from qudi.core.config import Configuration, ValidationError, YAMLError


def install_kernel():
    from jupyter_client.kernelspec import KernelSpecManager

    print('> Installing qudi kernel...')
    try:
        # prepare temporary kernelspec folder
        tempdir = tempfile.mkdtemp(suffix='_kernels')
        path = os.path.join(tempdir, 'qudi')
        kernel_path = os.path.abspath(__file__)
        os.mkdir(path)

        kernel_dict = {
            'argv': [sys.executable, kernel_path, '-f', '{connection_file}'],
            'display_name': 'qudi',
            'language': 'python'
        }
        # write the kernelspec file
        with open(os.path.join(path, 'kernel.json'), 'w') as f:
            json.dump(kernel_dict, f, indent=1)

        # install kernelspec folder
        kernel_spec_manager = KernelSpecManager()
        dest = kernel_spec_manager.install_kernel_spec(path, kernel_name='qudi', user=True)
        print(f'> Successfully installed kernelspec "qudi" in {dest}')
    finally:
        if os.path.isdir(tempdir):
            shutil.rmtree(tempdir)


def uninstall_kernel():
    from jupyter_client.kernelspec import KernelSpecManager

    print('> Uninstalling qudi kernel...')
    try:
        KernelSpecManager().remove_kernel_spec('qudi')
    except KeyError:
        print('> No kernelspec "qudi" found')
    else:
        print('> Successfully uninstalled kernelspec "qudi"')


class QudiKernelService(rpyc.Service):
    """
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._background_server = None

    def on_connect(self, conn):
        logging.warning(f'Qudi IPython kernel connected to local module service.')
        self._background_server = rpyc.BgServingThread(conn)

    def on_disconnect(self, conn):
        logging.warning(f'Qudi IPython kernel disconnected from local module service.')
        try:
            self._background_server.stop()
        except:
            pass
        finally:
            self._background_server = None

    # Implement methods starting with 'exposed_' here in order to provide services to qudi module
    # server.


class QudiKernelClient:
    """
    """

    def __init__(self):
        self.service_instance = QudiKernelService()
        self.connection = None

    def get_active_modules(self):
        if self.connection is None or self.connection.closed:
            return dict()
        try:
            return self.connection.root.get_namespace_dict()
        except (ConnectionError, EOFError):
            self.disconnect()
            return dict()

    def get_logger(self, name: str) -> logging.Logger:
        return self.connection.root.get_logger(name)

    def connect(self):
        config = Configuration()
        try:
            config.load()
        except (ValueError, ValidationError, YAMLError):
            pass
        self.connection = rpyc.connect(host='localhost',
                                       config={'allow_all_attrs': True,
                                               'allow_setattr': True,
                                               'allow_delattr': True,
                                               'allow_pickle': True,
                                               'sync_request_timeout': 3600},
                                       port=config['namespace_server_port'],
                                       service=self.service_instance)

    def disconnect(self):
        if self.connection is not None:
            try:
                self.connection.close()
            except:
                pass
            finally:
                self.connection = None


class QudiIPythonKernel(IPythonKernel):
    """
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._qudi_client = QudiKernelClient()
        self._qudi_client.connect()
        self._namespace_qudi_modules = set()
        self._qudi_logger = self._qudi_client.get_logger(f'QudiIPythonKernel_{str(self.ident)}')
        self.update_module_namespace()
        # Fixme: Dirty workaround after hours of searching on how to disable the insanely
        #  aggressive tab completion resolution of jedi that causes each descriptor (e.g. property)
        #  of the inspected object to be evaluated even if you do not want it to be inspected.
        #  Jupyter/IPython offer absolutely no documentation on anything not super shallow.
        # Fixme: Also tried to fix the crazy fomatting scheme of IPython this way. Clashes badly
        #  with rpyc otherwise.
        self.config.IPCompleter.use_jedi = False
        self.config.PlainTextFormatter.pprint = False
        # self.config.BaseFormatter.enabled = False
        # # lure out first warning and ignore
        # with warnings.catch_warnings():
        #     warnings.filterwarnings('ignore', module=r'traitlets', category=UserWarning)
        #     self.shell.run_cell('object()')

    def update_module_namespace(self):
        modules = self._qudi_client.get_active_modules()
        removed = self._namespace_qudi_modules.difference(modules)
        for mod in removed:
            self.shell.user_ns.pop(mod, None)
        self.shell.user_ns.pop('logger', None)
        self.shell.push({'logger': self._qudi_logger})
        self.shell.push(modules)
        self._namespace_qudi_modules = set(modules)

    # Update module namespace each time right before a cell is executed
    def do_execute(self, *args, **kwargs):
        self.update_module_namespace()
        return super().do_execute(*args, **kwargs)

    # Disconnect qudi remote module service before shutting down
    def do_shutdown(self, restart):
        self._qudi_client.disconnect()
        return super().do_shutdown(restart)


if __name__ == '__main__':
    if len(sys.argv) == 2 and sys.argv[1] == 'install':
        install_kernel()
    elif len(sys.argv) == 2 and sys.argv[1] == 'uninstall':
        uninstall_kernel()
    else:
        from ipykernel.kernelapp import IPKernelApp

        IPKernelApp.launch_instance(kernel_class=QudiIPythonKernel)



# -*- coding: utf-8 -*-
"""
This file contains utility methods to annotate arguments for which the user can potentially edit
values via GUI. These arguments are boiled down to simple builtin types that can be represented by
a GUI editor.

Parameters
----------
:int: :class:`~qudi.util.widgets.scientific_spinbox.ScienSpinBox`
    An integer type represented by a scientific spinbox widget.
:float: :class:`~qudi.util.widgets.scientific_spinbox.ScienDSpinbox`
    A floating-point type represented by a double precision scientific spinbox widget.
:str: :class:`~PySide2.QtWidgets.QLineEdit`
    A string type represented by a QLineEdit widget.
:complex: :class:`~qudi.util.widgets.literal_lineedit.ComplexLineEdit`
    A complex number type represented by a custom ComplexLineEdit widget.
:set: :class:`~qudi.util.widgets.literal_lineedit.SetLineEdit`
    A set type represented by a custom SetLineEdit widget.
:dict: :class:`~qudi.util.widgets.literal_lineedit.DictLineEdit`
    A dictionary type represented by a custom DictLineEdit widget.
:list: :class:`~qudi.util.widgets.literal_lineedit.ListLineEdit`
    A list type represented by a custom ListLineEdit widget.
:tuple: :class:`~qudi.util.widgets.literal_lineedit.TupleLineEdit`
    A tuple type represented by a custom TupleLineEdit widget.
:FilePath: :class:`~PySide2.QtWidgets.QLineEdit`
    A type for file paths represented by a QLineEdit widget.

Notes
-----
Here defined are also custom generic types that can be used like any other type from
the typing module.
Arguments annotated with these types are represented with the following widgets:
    FilePath: PySide2.QtWidgets.QLineEdit

Any other type annotation that can not be mapped to a simple builtin or a custom generic type
(see above) will result in no corresponding widget (None).

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['FilePath', 'RealNumber', 'ParameterWidgetMapper']

import inspect
import typing
from os import PathLike
from PySide2 import QtWidgets
from typing import Callable, Any, Set, FrozenSet, MutableSequence, Mapping, Tuple, Dict, Type, Union
from typing import get_origin, get_args, Iterable, Sequence

from qudi.util.helpers import is_complex_type, is_float_type, is_integer_type, is_string_type
from qudi.util.widgets.scientific_spinbox import ScienSpinBox, ScienDSpinBox
from qudi.util.widgets.literal_lineedit import ComplexLineEdit, TupleLineEdit, ListLineEdit
from qudi.util.widgets.literal_lineedit import SetLineEdit, DictLineEdit


FilePath = Union[str, bytes, PathLike]
RealNumber = Union[int, float]


class ParameterWidgetMapper:

    _type_widget_map = {int: ScienSpinBox,
                        float: ScienDSpinBox,
                        str: QtWidgets.QLineEdit,
                        PathLike: QtWidgets.QLineEdit,
                        complex: ComplexLineEdit,
                        set: SetLineEdit,
                        dict: DictLineEdit,
                        tuple: TupleLineEdit,
                        list: ListLineEdit}

    @classmethod
    def widgets_for_callable(cls, func: Callable) -> Dict[str, Union[Type[QtWidgets.QWidget], None]]:
        """Returns QWidget classes for each parameter from the call signature of "func".
        See ParameterWidgetMapper.widget_for_parameter for more information.
        """
        sig = inspect.signature(func)
        return {name: cls.widget_for_parameter(param) for name, param in sig.parameters.items()}

    @classmethod
    def widget_for_parameter(cls, param: inspect.Parameter) -> Union[Type[QtWidgets.QWidget], None]:
        """Tries to determine a suitable QWidget to represent the given parameter.
        If no type annotation is given for a parameter it will try to determine the type from the
        default value.
        If no (known) type can be determined, None will be returned.
        """
        # Try to deduce missing type annotation from default parameter type
        if param.annotation is inspect.Parameter.empty:
            if param.default is inspect.Parameter.empty:
                return None
            else:
                return cls.widget_from_value(param.default)
        else:
            return cls.widget_from_annotation(param.annotation)

    @classmethod
    def widget_from_value(cls, value: Any) -> Union[Type[QtWidgets.QWidget], None]:
        """Tries to determine a suitable QWidget to represent the type of the given value.
        """
        normalized_type = cls._normalize_type(type(value))
        return cls._type_widget_map.get(normalized_type, None)

    @classmethod
    def widget_from_annotation(cls, annotation: Any) -> Union[Type[QtWidgets.QWidget], None]:
        """Tries to determine a suitable QWidget to represent values of the given annotation type.
        """
        normalized_type = cls._annotation_to_type(annotation)
        return cls._type_widget_map.get(normalized_type, None)

    @staticmethod
    def _normalize_type(typ: Type) -> Type:
        """Normalizes given type to a base/builtin type.
        Examples:
                      numpy.float32 -> float
            collections.OrderedDict -> dict
        """
        if is_string_type(typ):
            return str
        elif is_integer_type(typ):
            return int
        elif is_float_type(typ):
            return float
        elif is_complex_type(typ):
            return complex
        elif issubclass(typ, (bytes, PathLike)):
            return PathLike
        elif issubclass(typ, (Set, FrozenSet)):
            return set
        elif issubclass(typ, Mapping):
            return dict
        elif issubclass(typ, MutableSequence):
            return list
        elif issubclass(typ, (Tuple, Iterable, Sequence)):
            return tuple
        return None

    @classmethod
    def _annotation_to_type(cls, annotation: Any) -> Type:
        """Converts a type annotation (e.g. from a callable signature) to a normalized type.
        See ParameterWidgetMapper._normalize_type for more information.
        """
        # If annotation is optional, call this method again on the first argument in order to get a
        # type
        if cls._is_optional_annotation(annotation):
            return cls._annotation_to_type(get_args(annotation)[0])

        if annotation == RealNumber:
            return float
        elif annotation == FilePath:
            return PathLike
        else:
            try:
                if inspect.isclass(annotation):
                    return cls._normalize_type(annotation)
                else:
                    return cls._normalize_type(get_origin(annotation))
            except TypeError:
                pass
        return None

    @staticmethod
    def _is_optional_annotation(annotation: Any) -> bool:
        """Check if an annotation is optional, i.e. if it is typing.Union with two arguments,
        the second one being NoneType.
        """
        if get_origin(annotation) == typing.Union:
            args = get_args(annotation)
            return len(args) == 2 and issubclass(args[1], type(None))
        return False



# -*- coding: utf-8 -*-

"""
This file contains models of Lorentzian fitting routines for qudi based on the lmfit package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ComplexLorentzian', 'DoubleLorentzian', 'Lorentzian', 'LorentzianLinear',
           'TripleLorentzian', 'multiple_lorentzian', 'multiple_complex_lorentzian']

import numpy as np
from typing import Sequence
from qudi.util.fit_models.model import FitModelBase, estimator
from qudi.util.fit_models.helpers import correct_offset_histogram, smooth_data, sort_check_data
from qudi.util.fit_models.helpers import estimate_double_peaks, estimate_triple_peaks
from qudi.util.fit_models.linear import Linear


def multiple_lorentzian(x, centers, sigmas, amplitudes):
    """ Mathematical definition of the sum of multiple (physical) Lorentzian functions without any
    bias.

    WARNING: Iterable parameters "centers", "sigmas", and "amplitudes" must have the same length.

    Parameters
    ----------
    x : float
        The independent variable to calculate lorentz(x).
    centers : iterable
        Iterable containing center positions for all Lorentzians.
    sigmas : iterable
        Iterable containing width parameters (half-width at half-maximum) for all Lorentzians.
    amplitudes : iterable
        Iterable containing amplitudes for all Lorentzians.

    Returns
    -------
    float
        The result given x for lorentz(x).

    """
    assert len(centers) == len(sigmas) == len(amplitudes)
    return sum(amp * sig ** 2 / ((x - c) ** 2 + sig ** 2) for c, sig, amp in
               zip(centers, sigmas, amplitudes))


def multiple_complex_lorentzian(x: float, centers: Sequence[float], sigmas: Sequence[float],
                                amplitudes: Sequence[float], thetas: Sequence[float]):
    """ Mathematical definition of the sum of multiple complex Lorentzian functions without any
    bias.

    WARNING: Values in "thetas" must be in deg (NOT rad).

    WARNING: Sequence parameters "centers", "sigmas", "amplitudes" and "thetas" must have same
    length.
    """
    assert len(centers) == len(sigmas) == len(amplitudes) == len(thetas)
    return sum(np.exp(1j * np.deg2rad(th)) * amp * (-1j * (x - c) * sig + sig ** 2) / (
                np.pi * sig * ((x - c) ** 2 + sig ** 2)) for c, sig, amp, th in
               zip(centers, sigmas, amplitudes, thetas))


class Lorentzian(FitModelBase):
    """
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=0., min=0., max=np.inf)
        self.set_param_hint('center', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('sigma', value=0., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, center, sigma, amplitude):
        return offset + multiple_lorentzian(x, (center,), (sigma,), (amplitude,))

    @estimator('Peak')
    def estimate_peak(self, data, x):
        data, x = sort_check_data(data, x)
        # Smooth data
        filter_width = max(1, int(round(len(x) / 20)))
        data_smoothed, _ = smooth_data(data, filter_width)
        data_smoothed, offset = correct_offset_histogram(data_smoothed, bin_width=2 * filter_width)

        # determine peak position
        center = x[np.argmax(data_smoothed)]

        # calculate amplitude
        amplitude = abs(max(data_smoothed))

        # according to the derived formula, calculate sigma. The crucial part is here that the
        # offset was estimated correctly, then the area under the curve is calculated correctly:
        numerical_integral = np.trapz(data_smoothed, x)
        sigma = abs(numerical_integral / (np.pi * amplitude))

        x_spacing = min(abs(np.ediff1d(x)))
        x_span = abs(x[-1] - x[0])
        data_span = abs(max(data) - min(data))

        estimate = self.make_params()
        estimate['amplitude'].set(value=amplitude, min=0, max=2 * amplitude)
        estimate['sigma'].set(value=sigma, min=x_spacing, max=x_span)
        estimate['center'].set(value=center, min=min(x) - x_span / 2, max=max(x) + x_span / 2)
        estimate['offset'].set(
            value=offset, min=min(data) - data_span / 2, max=max(data) + data_span / 2
        )
        return estimate

    @estimator('Dip')
    def estimate_dip(self, data, x):
        estimate = self.estimate_peak(-data, x)
        estimate['offset'].set(value=-estimate['offset'].value,
                               min=-estimate['offset'].max,
                               max=-estimate['offset'].min)
        estimate['amplitude'].set(value=-estimate['amplitude'].value,
                                  min=-estimate['amplitude'].max,
                                  max=-estimate['amplitude'].min)
        return estimate

    @estimator('Peak (no offset)')
    def estimate_peak_no_offset(self, data, x):
        estimate = self.estimate_peak(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Dip (no offset)')
    def estimate_dip_no_offset(self, data, x):
        estimate = self.estimate_dip(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate


class DoubleLorentzian(FitModelBase):
    """ ToDo: Document
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_2', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('center_1', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('center_2', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('sigma_1', value=0., min=0., max=np.inf)
        self.set_param_hint('sigma_2', value=0., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, center_1, center_2, sigma_1, sigma_2, amplitude_1, amplitude_2):
        return offset + multiple_lorentzian(x,
                                                (center_1, center_2),
                                                (sigma_1, sigma_2),
                                                (amplitude_1, amplitude_2))

    @estimator('Peaks')
    def estimate_peaks(self, data, x):
        data, x = sort_check_data(data, x)
        data_smoothed, filter_width = smooth_data(data)
        leveled_data_smooth, offset = correct_offset_histogram(data_smoothed,
                                                               bin_width=2 * filter_width)
        estimate, limits = estimate_double_peaks(leveled_data_smooth, x, filter_width)

        params = self.make_params()
        params['amplitude_1'].set(value=estimate['height'][0],
                                  min=limits['height'][0][0],
                                  max=limits['height'][0][1])
        params['amplitude_2'].set(value=estimate['height'][1],
                                  min=limits['height'][1][0],
                                  max=limits['height'][1][1])
        params['center_1'].set(value=estimate['center'][0],
                               min=limits['center'][0][0],
                               max=limits['center'][0][1])
        params['center_2'].set(value=estimate['center'][1],
                               min=limits['center'][1][0],
                               max=limits['center'][1][1])
        params['sigma_1'].set(value=estimate['fwhm'][0] / 2.3548,
                              min=limits['fwhm'][0][0] / 2.3548,
                              max=limits['fwhm'][0][1] / 2.3548)
        params['sigma_2'].set(value=estimate['fwhm'][1] / 2.3548,
                              min=limits['fwhm'][1][0] / 2.3548,
                              max=limits['fwhm'][1][1] / 2.3548)
        return params

    @estimator('Dips')
    def estimate_dips(self, data, x):
        estimate = self.estimate_peaks(-data, x)
        estimate['offset'].set(value=-estimate['offset'].value,
                               min=-estimate['offset'].max,
                               max=-estimate['offset'].min)
        estimate['amplitude_1'].set(value=-estimate['amplitude_1'].value,
                                    min=-estimate['amplitude_1'].max,
                                    max=-estimate['amplitude_1'].min)
        estimate['amplitude_2'].set(value=-estimate['amplitude_2'].value,
                                    min=-estimate['amplitude_2'].max,
                                    max=-estimate['amplitude_2'].min)
        return estimate


class TripleLorentzian(FitModelBase):
    """ ToDo: Document
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_2', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_3', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('center_1', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('center_2', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('center_3', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('sigma_1', value=0., min=0., max=np.inf)
        self.set_param_hint('sigma_2', value=0., min=0., max=np.inf)
        self.set_param_hint('sigma_3', value=0., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, center_1, center_2, center_3, sigma_1, sigma_2, sigma_3,
                        amplitude_1, amplitude_2, amplitude_3):
        return offset + multiple_lorentzian(x,
                                            (center_1, center_2, center_3),
                                            (sigma_1, sigma_2, sigma_3),
                                            (amplitude_1, amplitude_2, amplitude_3))

    @estimator('Peaks')
    def estimate_peaks(self, data, x):
        data, x = sort_check_data(data, x)
        data_smoothed, filter_width = smooth_data(data)
        leveled_data_smooth, offset = correct_offset_histogram(data_smoothed,
                                                               bin_width=2 * filter_width)
        estimate, limits = estimate_triple_peaks(leveled_data_smooth, x, filter_width)

        params = self.make_params()
        params['amplitude_1'].set(value=estimate['height'][0],
                                  min=limits['height'][0][0],
                                  max=limits['height'][0][1])
        params['amplitude_2'].set(value=estimate['height'][1],
                                  min=limits['height'][1][0],
                                  max=limits['height'][1][1])
        params['amplitude_3'].set(value=estimate['height'][2],
                                  min=limits['height'][2][0],
                                  max=limits['height'][2][1])
        params['center_1'].set(value=estimate['center'][0],
                               min=limits['center'][0][0],
                               max=limits['center'][0][1])
        params['center_2'].set(value=estimate['center'][1],
                               min=limits['center'][1][0],
                               max=limits['center'][1][1])
        params['center_3'].set(value=estimate['center'][2],
                               min=limits['center'][2][0],
                               max=limits['center'][2][1])
        params['sigma_1'].set(value=estimate['fwhm'][0] / 2.3548,
                              min=limits['fwhm'][0][0] / 2.3548,
                              max=limits['fwhm'][0][1] / 2.3548)
        params['sigma_2'].set(value=estimate['fwhm'][1] / 2.3548,
                              min=limits['fwhm'][1][0] / 2.3548,
                              max=limits['fwhm'][1][1] / 2.3548)
        params['sigma_3'].set(value=estimate['fwhm'][2] / 2.3548,
                              min=limits['fwhm'][2][0] / 2.3548,
                              max=limits['fwhm'][2][1] / 2.3548)
        return params

    @estimator('Dips')
    def estimate_dips(self, data, x):
        estimate = self.estimate_peaks(-data, x)
        estimate['offset'].set(value=-estimate['offset'].value,
                               min=-estimate['offset'].max,
                               max=-estimate['offset'].min)
        estimate['amplitude_1'].set(value=-estimate['amplitude_1'].value,
                                    min=-estimate['amplitude_1'].max,
                                    max=-estimate['amplitude_1'].min)
        estimate['amplitude_2'].set(value=-estimate['amplitude_2'].value,
                                    min=-estimate['amplitude_2'].max,
                                    max=-estimate['amplitude_2'].min)
        estimate['amplitude_3'].set(value=-estimate['amplitude_3'].value,
                                    min=-estimate['amplitude_3'].max,
                                    max=-estimate['amplitude_3'].min)
        return estimate


class LorentzianLinear(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('slope', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=0, min=-np.inf, max=np.inf)
        self.set_param_hint('center', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('sigma', value=0., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, slope, center, sigma, amplitude):
        x0 = (x - min(x))
        return offset + x0 * slope + multiple_lorentzian(x, (center,), (sigma,), (amplitude,))

    @estimator('Peak')
    def estimate_peak(self, data, x):
        data, x = sort_check_data(data, x)
        data_span = abs(max(data) - min(data))

        # Perform a normal Lorentzian peak fit and subtract the result from data
        model = Lorentzian()
        gauss_fit = model.fit(data, model.estimate_peak(data, x), x=x)
        data_sub = data - gauss_fit.best_fit
        # Perform a linear fit in subtracted data in order to estimate slope
        model = Linear()
        linear_fit = model.fit(data_sub, model.estimate(data_sub, x), x=x)
        offset = linear_fit.params['offset'].value + min(x) * linear_fit.params['slope'].value

        # Merge fit results into parameter estimates
        estimate = self.make_params()
        estimate['offset'].set(value=offset,
                               min=min(data) - data_span / 2,
                               max=max(data) + data_span / 2,
                               vary=True)
        estimate['slope'].set(value=linear_fit.params['slope'].value, min=-np.inf, max=np.inf, vary=True)
        estimate['amplitude'].set(value=gauss_fit.params['amplitude'].value,
                                  min=gauss_fit.params['amplitude'].min,
                                  max=gauss_fit.params['amplitude'].max,
                                  vary=True)
        estimate['center'].set(value=gauss_fit.params['center'].value,
                               min=gauss_fit.params['center'].min,
                               max=gauss_fit.params['center'].max,
                               vary=True)
        estimate['sigma'].set(value=gauss_fit.params['sigma'].value,
                              min=gauss_fit.params['sigma'].min,
                              max=gauss_fit.params['sigma'].max,
                              vary=True)
        return estimate

    @estimator('Dip')
    def estimate_dip(self, data, x):
        estimate = self.estimate_peak(-data, x)
        estimate['offset'].set(value=-estimate['offset'].value,
                               min=-estimate['offset'].max,
                               max=-estimate['offset'].min)
        estimate['amplitude'].set(value=-estimate['amplitude'].value,
                                  min=-estimate['amplitude'].max,
                                  max=-estimate['amplitude'].min)
        return estimate


class ComplexLorentzian(FitModelBase):
    """ ToDo: Implement estimators
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('amplitude', value=0., min=0., max=np.inf)
        self.set_param_hint('center', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('sigma', value=1e-15, min=0, max=np.inf)
        self.set_param_hint('theta', value=0., min=-180., max=180.)

    @staticmethod
    def _model_function(x, center, sigma, amplitude, theta):
        return multiple_complex_lorentzian(x, (center,), (sigma,), (amplitude,), (theta,))



# -*- coding: utf-8 -*-
"""
This file extends the ruamel.yaml package functionality to load and dump more data types needed by
qudi (mostly numpy array and number types).
Provides easy to use yaml_load and yaml_dump functions to read and write qudi YAML files.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['SafeRepresenter', 'SafeConstructor', 'YAML', 'yaml_load', 'yaml_dump', 'ParserError',
           'YAMLError', 'MarkedYAMLError', 'YAMLStreamError', 'ScannerError', 'ConstructorError',
           'DuplicateKeyError']

import os
import numpy as np
import ruamel.yaml as _yaml
from ruamel.yaml.error import YAMLError, MarkedYAMLError, YAMLStreamError
from ruamel.yaml.parser import ParserError, ScannerError
from ruamel.yaml.constructor import ConstructorError, DuplicateKeyError
from enum import Enum, IntEnum, IntFlag, Flag
from importlib import import_module
from collections import OrderedDict
from io import BytesIO, TextIOWrapper
from typing import Optional, Any, Mapping, Dict, Union


_FilePath = Union[str, bytes, os.PathLike]


class SafeRepresenter(_yaml.SafeRepresenter):
    """Custom YAML representer for qudi config files.
    """
    ndarray_max_size = 20

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._extndarray_count = 0

    def ignore_aliases(self, ignore_data):
        """Ignore aliases and anchors. Overwrites base class implementation.
        """
        return True

    def represent_numpy_int(self, data):
        """Representer for numpy int scalars.
        """
        return self.represent_int(data.item())

    def represent_numpy_float(self, data):
        """Representer for numpy float scalars.
        """
        return self.represent_float(data.item())

    def represent_numpy_complex(self, data):
        """Representer for numpy complex scalars.
        """
        return self.represent_complex(data.item())

    def represent_dict_no_sort(self, data):
        """Representer for dict and OrderedDict to prevent ruamel.yaml from sorting keys.
        """
        return self.represent_dict(data.items())

    def represent_complex(self, data):
        """Representer for builtin complex type.
        """
        return self.represent_scalar(tag='tag:yaml.org,2002:complex', value=str(data))

    def represent_frozenset(self, data):
        """Representer for builtin frozenset type.
        """
        node = self.represent_set(data)
        node.tag = 'tag:yaml.org,2002:frozenset'
        return node

    def represent_enum(self, data):
        """Representer for enum types with base class enum.
        """
        class_name = data.__class__.__name__
        module = data.__class__.__module__
        try:
            mod = import_module(module)
            cls = getattr(mod, class_name)
            assert data == cls[data.name]
        except (AttributeError, ImportError, AssertionError):
            raise TypeError(f'Data can not be represented as enum.Enum.')
        return self.represent_scalar(tag='tag:yaml.org,2002:enum',
                                     value=f'{module}.{class_name}[{data.name}]')

    def represent_flag(self, data):
        """Representer for enum types with base class enum.
        """
        class_name = data.__class__.__name__
        module = data.__class__.__module__
        try:
            mod = import_module(module)
            cls = getattr(mod, class_name)
            assert data == cls(data.value)
        except (AttributeError, ImportError, AssertionError):
            raise TypeError(f'Data can not be represented as enum.Flag')
        return self.represent_scalar(tag='tag:yaml.org,2002:flag',
                                     value=f'{module}.{class_name}({data.value:d})')

    def represent_ndarray(self, data):
        """Representer for numpy.ndarrays.
        Will represent the array in binary representation as ASCII-encoded string by default.
        If the output stream to dump to is a "regular" open text file handle (io.TextIOWrapper) and
        the array size exceeds the specified maximum ndarray size, it is dumped into a separate
        binary .npy file and is represented in YAML as file path string.
        """
        # Write to separate file if possible and required (array size > self.ndarray_max_size)
        # FIXME: Find a better way... this is a mean hack to get the file path to dump,
        if isinstance(self.dumper._output, TextIOWrapper) and data.size > self.ndarray_max_size:
            try:
                out_stream_path = self.dumper._output.name
                dir_path = os.path.dirname(out_stream_path)
                file_name = os.path.splitext(os.path.basename(out_stream_path))[0]
                file_path = f'{os.path.join(dir_path, file_name)}-{self._extndarray_count:06}.npy'
                np.save(file_path, data, allow_pickle=False, fix_imports=False)
                self._extndarray_count += 1
                return self.represent_scalar(tag='tag:yaml.org,2002:extndarray', value=file_path)
            except:
                pass

        # Represent as binary stream (ASCII-encoded) by default
        with BytesIO() as f:
            np.save(f, data, allow_pickle=False, fix_imports=False)
            binary_repr = f.getvalue()
        node = self.represent_binary(binary_repr)
        node.tag = 'tag:yaml.org,2002:ndarray'
        return node


# register custom representers
SafeRepresenter.add_representer(frozenset, SafeRepresenter.represent_frozenset)
SafeRepresenter.add_representer(complex, SafeRepresenter.represent_complex)
SafeRepresenter.add_representer(dict, SafeRepresenter.represent_dict_no_sort)
SafeRepresenter.add_representer(OrderedDict, SafeRepresenter.represent_dict_no_sort)
SafeRepresenter.add_representer(np.ndarray, SafeRepresenter.represent_ndarray)
SafeRepresenter.add_multi_representer(Enum, SafeRepresenter.represent_enum)
SafeRepresenter.add_multi_representer(IntEnum, SafeRepresenter.represent_enum)
SafeRepresenter.add_multi_representer(Flag, SafeRepresenter.represent_flag)
SafeRepresenter.add_multi_representer(IntFlag, SafeRepresenter.represent_flag)
SafeRepresenter.add_multi_representer(np.integer, SafeRepresenter.represent_numpy_int)
SafeRepresenter.add_multi_representer(np.floating, SafeRepresenter.represent_numpy_float)
SafeRepresenter.add_multi_representer(np.complexfloating, SafeRepresenter.represent_numpy_complex)


class SafeConstructor(_yaml.SafeConstructor):
    """Custom YAML constructor for qudi config files.
    """

    def construct_ndarray(self, node):
        """The constructor for a numpy array that is saved as binary string with ASCII-encoding.
        """
        value = self.construct_yaml_binary(node)
        with BytesIO(value) as f:
            return np.load(f)

    def construct_extndarray(self, node):
        """The constructor for a numpy array that is saved in a separate file.
        """
        return np.load(self.construct_yaml_str(node), allow_pickle=False, fix_imports=False)

    def construct_frozenset(self, node):
        """The frozenset constructor.
        """
        try:
            # FIXME: The returned generator does not properly work with iteration using next()
            return frozenset(tuple(self.construct_yaml_set(node))[0])
        except IndexError:
            return frozenset()

    def construct_complex(self, node):
        """The complex constructor.
        """
        return complex(self.construct_yaml_str(node))

    def construct_enum(self, node):
        """The Enum constructor.
        """
        enum_repr_str = self.construct_yaml_str(node)
        enum_mod_cls, enum_name = enum_repr_str.rsplit(']', 1)[0].rsplit('[', 1)
        module, cls_name = enum_mod_cls.rsplit('.', 1)
        cls = getattr(import_module(module), cls_name)
        return cls[enum_name]

    def construct_flag(self, node):
        """The Flag constructor.
        """
        enum_repr_str = self.construct_yaml_str(node)
        enum_mod_cls, enum_value_str = enum_repr_str.rsplit(')', 1)[0].rsplit('(', 1)
        module, cls_name = enum_mod_cls.rsplit('.', 1)
        cls = getattr(import_module(module), cls_name)
        return cls(int(enum_value_str))


# register custom constructors
SafeConstructor.add_constructor('tag:yaml.org,2002:frozenset', SafeConstructor.construct_frozenset)
SafeConstructor.add_constructor('tag:yaml.org,2002:complex', SafeConstructor.construct_complex)
SafeConstructor.add_constructor('tag:yaml.org,2002:ndarray', SafeConstructor.construct_ndarray)
SafeConstructor.add_constructor('tag:yaml.org,2002:extndarray',
                                SafeConstructor.construct_extndarray)
SafeConstructor.add_constructor('tag:yaml.org,2002:enum', SafeConstructor.construct_enum)
SafeConstructor.add_constructor('tag:yaml.org,2002:flag', SafeConstructor.construct_flag)


class YAML(_yaml.YAML):
    """ruamel.yaml.YAML subclass to be used by qudi for all loading/dumping purposes.
    Will always use the 'safe' option without round-trip functionality.
    """

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        kwargs
            Keyword arguments accepted by ruamel.yaml.YAML(), excluding "typ".
        """
        kwargs['typ'] = 'safe'
        super().__init__(**kwargs)
        self.default_flow_style = False
        self.Representer = SafeRepresenter
        self.Constructor = SafeConstructor


def yaml_load(file_path: _FilePath, ignore_missing: Optional[bool] = False) -> Dict[str, Any]:
    """Loads a qudi style YAML file.
    Raises OSError if the file does not exist or can not be accessed.

    Parameters
    ----------
    file_path : str
        Path to config file.
    ignore_missing : bool, optional
        Flag to suppress FileNotFoundError.

    Returns
    -------
    dict
        The data as python/numpy objects in a dict.
    """
    try:
        with open(file_path, 'r') as f:
            data = YAML().load(f)
            # yaml returns None if the stream was empty
            return dict() if data is None else data
    except OSError:
        if ignore_missing:
            return dict()
        else:
            raise


def yaml_dump(file_path: _FilePath, data: Mapping[str, Any]) -> None:
    """Saves data to file_path in qudi style YAML format. Creates subdirectories if needed.

    file_path : str
        Path to YAML file to save data into.
    data : dict
        Dict containing the data to save to file.
    """
    file_dir = os.path.dirname(file_path)
    if file_dir:
        os.makedirs(file_dir, exist_ok=True)
    with open(file_path, 'w') as f:
        YAML().dump(data, f)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.

ToDo: Throw errors around for non-existent directories
"""

__all__ = ['get_appdata_dir', 'get_default_config_dir', 'get_default_log_dir',
           'get_default_data_dir', 'get_daily_directory', 'get_home_dir', 'get_main_dir',
           'get_userdata_dir', 'get_artwork_dir', 'get_module_app_data_path']

import datetime
import os
import sys
from typing import Optional


def get_main_dir() -> str:
    """
    Returns the absolute path to the directory of the main software.

    Returns
    -------
    str
        Path to the main tree of the software.
    """
    import qudi.core as core
    return os.path.abspath(os.path.join(os.path.dirname(core.__file__), '..'))


def get_artwork_dir() -> str:
    """
    Returns the absolute path to the Qudi artwork directory.

    Returns
    -------
    str
        Path to the artwork directory of Qudi.
    """
    return os.path.join(get_main_dir(), 'artwork')


def get_home_dir() -> str:
    """
    Returns the absolute path to the home directory.

    Returns
    -------
    str
        Absolute path to the home directory.
    """
    return os.path.abspath(os.path.expanduser('~'))


def get_userdata_dir(create_missing: Optional[bool] = False) -> str:
    """
    Returns the absolute path to the Qudi subfolder in the user home directory.
    This path should be used for exposed user data like config files, etc.

    Returns
    -------
    str
        Absolute path to the Qudi subfolder in the user home directory.
    """
    path = os.path.join(get_home_dir(), 'qudi')
    # Create directory if desired. Will throw an exception if path returned by get_home_dir() is
    # non-existent (which should never happen).
    if create_missing and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_appdata_dir(create_missing: Optional[bool] = False) -> str:
    """
    Get the system-specific application data directory.

    Returns
    -------
    str
        Path to the application data directory specific to the system.
    """
    if sys.platform == 'win32':
        # resolves to "C:\Documents and Settings\<UserName>\Application Data" on XP and
        # "C:\Users\<UserName>\AppData\Roaming" on win7 and newer
        path = os.path.join(os.environ['APPDATA'], 'qudi')
    elif sys.platform == 'darwin':
        path = os.path.abspath(os.path.expanduser('~/Library/Preferences/qudi'))
    else:
        path = os.path.abspath(os.path.expanduser('~/.local/qudi'))

    # Create path if desired.
    if create_missing and not os.path.exists(path):
        os.makedirs(path)
    return path


def get_default_config_dir(create_missing: Optional[bool] = False) -> str:
    """
    Get the system-specific application data directory.

    Returns
    -------
    str
        Path to the application data directory specific to the system.

    """
    path = os.path.join(get_userdata_dir(create_missing), 'config')
    # Create path if desired.
    if create_missing and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_default_log_dir(create_missing: Optional[bool] = False) -> str:
    """
    Get the system-specific application log directory.

    Returns
    -------
    str
        Path to the default logging directory specific to the system.

    """
    # FIXME: This needs to be properly done for linux systems
    path = os.path.join(get_userdata_dir(create_missing), 'log')
    # Create path if desired.
    if create_missing and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_default_data_dir(create_missing: Optional[bool] = False) -> str:
    """Get the system specific application fallback data root directory.
    Does NOT consider qudi configuration.

    Returns
    -------
    str
        Path to default data root directory.
    """
    # FIXME: This needs to be properly done for linux systems
    path = os.path.join(get_userdata_dir(create_missing), 'Data')
    # Create path if desired.
    if create_missing and not os.path.exists(path):
        os.mkdir(path)
    return path


def get_daily_directory(timestamp: Optional[datetime.datetime] = None, root: Optional[str] = None,
                        create_missing: Optional[bool] = False) -> str:
    """
    Returns a path tree according to the timestamp given.

    The directory structure will have the form: root/<YYYY>/<MM>/<YYYY-MM-DD>
    If no root directory is given, this method will return just the relative path stub:
    <YYYY>/<MM>/<YYYY-MM-DD>

    Parameters
    ----------
    timestamp : datetime.datetime, optional
        Timestamp for which to create the daily directory. Defaults to current timestamp if not provided.
    root : str, optional
        Root path for the daily directory structure. If not provided, only the relative path stub is returned.
    create_missing : bool, optional
        Indicates if the directory should be created (True) or not (False). Only considered if root is given.

    Returns
    -------
    str
        Path representing the directory structure based on the timestamp.
    """
    if timestamp is None:
        timestamp = datetime.datetime.now()

    day_dir = timestamp.strftime('%Y-%m-%d')
    year_dir, month_dir = day_dir.split('-')[:2]
    daily_path = os.path.join(year_dir, month_dir, day_dir)
    if root is not None:
        daily_path = os.path.join(root, daily_path)
        if create_missing:
            os.makedirs(daily_path, exist_ok=True)
    return daily_path


def get_module_app_data_path(cls_name: str, module_base: str, module_name: str) -> str:
    """Constructs the appData file path for the given qudi module.
    """
    file_name = f'status-{cls_name}_{module_base}_{module_name}.cfg'
    return os.path.join(get_appdata_dir(), file_name)



# -*- coding: utf-8 -*-

"""
This file contains unit tests for all qudi fit routines for Lorentzian peak/dip models.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import unittest
import numpy as np

from qudi.util.fit_models.lorentzian import Lorentzian


class TestLorentzianMethods(unittest.TestCase):
    _fit_param_tolerance = 0.05  # 5% tolerance for each fit parameter

    @staticmethod
    def lorentzian(x, offset, amplitude, center, sigma):
        return offset + amplitude * sigma ** 2 / ((x - center) ** 2 + sigma ** 2)

    def setUp(self):
        self.offset = (np.random.rand() - 0.5) * 2e6
        self.amplitude = np.random.rand() * 100
        window = max(1e-9, np.random.rand() * 1e9)
        left = (np.random.rand() - 0.5) * 2e9
        points = np.random.randint(10, 1001)
        self.x_values = np.linspace(left, left + window, points)
        min_sigma = 1.5 * (self.x_values[1] - self.x_values[0])
        self.sigma = min_sigma + np.random.rand() * ((window / 2) - min_sigma)
        self.center = left + np.random.rand() * window
        self.noise_amp = max(self.amplitude / 10, np.random.rand() * self.amplitude)
        self.noise = (np.random.rand(points) - 0.5) * self.noise_amp

    def test_gaussian(self):
        # Test for lorentzian peak
        y_values = self.noise + self.lorentzian(self.x_values,
                                                self.offset,
                                                self.amplitude,
                                                self.center,
                                                self.sigma)

        fit_model = Lorentzian()
        fit_result = fit_model.fit(data=y_values,
                                   x=self.x_values,
                                   **fit_model.guess(y_values, self.x_values))

        params_ideal = {'offset': self.offset,
                        'amplitude': self.amplitude,
                        'center': self.center,
                        'sigma': self.sigma}
        for name, fit_param in fit_result.best_values.items():
            diff = abs(fit_param - params_ideal[name])
            tolerance = abs(params_ideal[name] * self._fit_param_tolerance)
            msg = 'Lorentzian peak fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)

        # Test for lorentzian dip
        y_values = self.noise + self.lorentzian(self.x_values,
                                                self.offset,
                                                -self.amplitude,
                                                self.center,
                                                self.sigma)

        fit_model = Lorentzian()
        fit_result = fit_model.fit(data=y_values,
                                   x=self.x_values,
                                   **fit_model.guess(y_values, self.x_values))

        params_ideal['amplitude'] = -self.amplitude
        for name, fit_param in fit_result.best_values.items():
            diff = abs(fit_param - params_ideal[name])
            tolerance = abs(params_ideal[name] * self._fit_param_tolerance)
            msg = 'Lorentzian dip fit parameter "{0}" not within {1:.2%} tolerance'.format(
                name, self._fit_param_tolerance
            )
            self.assertLessEqual(diff, tolerance, msg)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
This file contains a touch-like toggle switch.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ToggleSwitch']

from PySide2 import QtWidgets, QtCore, QtGui


class ToggleSwitch(QtWidgets.QAbstractButton):
    """ A mobile/touch inspired toggle switch to switch between two states.
    """

    sigStateChanged = QtCore.Signal(str)

    def __init__(self, parent=None, state_names=None, thumb_track_ratio=1, scale_text=True,
                 display_text=True):
        super().__init__(parent=parent)
        self.setCheckable(True)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        # check state_names
        if state_names is None:
            self._state_names = None
        elif len(state_names) != 2 or not all(isinstance(n, str) and n != '' for n in state_names):
            raise ValueError(
                f'state_names must be None or sequence of exactly 2 non-empty strings. '
                f'Received: {state_names}'
            )
        else:
            self._state_names = tuple(state_names)
        # check thumb_track_ratio
        if thumb_track_ratio <= 0:
            raise ValueError(
                f'thumb_track_ratio must have a value > 0. Received: {thumb_track_ratio}'
            )
        self._thumb_track_ratio = thumb_track_ratio
        # check scale_text
        if not isinstance(scale_text, bool):
            raise ValueError(f'scale_text must be bool type. Received: {scale_text}')
        self._scale_text = scale_text
        # check display_text
        if not isinstance(display_text, bool):
            raise ValueError(f'display_text must be bool type. Received: {display_text}')
        self._display_text = display_text if self._state_names else False
        self.__display_text = self._display_text

        # Determine (minimal) size hint based on text to display
        self._default_text_size = None
        self._size_hint = None
        self._refresh_size_hint()

        # Calculate geometry for painting
        self._thumb_radius = 0
        self._track_radius = 0
        self._track_margin = 0
        self._thumb_origin = 0
        self._current_text_width = 0
        self._refresh_geometry()
        self._refresh_text_scale()

        # Determine appearance from current palette depending on thumb style
        palette = self.palette()
        if self._thumb_track_ratio > 1:
            self._track_colors = (palette.dark(), palette.highlight())
            self._thumb_colors = (palette.light(), palette.highlight())
            self._text_colors = (palette.text().color(), palette.highlightedText().color())
            self._track_opacity = 0.5
        else:
            self._track_colors = (palette.dark(), palette.highlight())
            self._thumb_colors = (palette.light(), palette.highlightedText())
            self._text_colors = (palette.text().color(), palette.highlightedText().color())
            self._track_opacity = 1

        # property value for current thumb position
        self._thumb_position = self._thumb_origin

        # Connect notifier signal
        self.clicked.connect(self._notify_state_change)

        # set up the animation
        self._slider_animation = QtCore.QPropertyAnimation(self, b'thumb_position', self)
        self._slider_animation.finished.connect(self._finish_animation)

    @QtCore.Slot()
    def _finish_animation(self):
        target_position = self._thumb_end
        if self._thumb_position != target_position:
            self.thumb_position = target_position

    def _refresh_size_hint(self):
        metrics = QtGui.QFontMetrics(self.font())
        if self._display_text:
            self._default_text_size = QtCore.QSize(
                max(metrics.horizontalAdvance(f' {text} ') for text in self._state_names),
                metrics.height()
            )
        else:
            self._default_text_size = QtCore.QSize(metrics.horizontalAdvance(' OFF '),
                                                   metrics.height())
        if self._thumb_track_ratio <= 1:
            height = self._default_text_size.height() * 1.5
        else:
            height = self._default_text_size.height() * 1.5 * self._thumb_track_ratio
        width = self._default_text_size.width() + 2 * height
        self._size_hint = QtCore.QSize(width, height)
        self.setMinimumSize(self._size_hint)

    def _refresh_text_scale(self):
        if not self._display_text:
            self._current_text_width = 0
            self.__display_text = False
            return

        if self._scale_text:
            # Determine current maximum height and width for text field
            max_height = int(round(1.5 * self._track_radius))
            if self._thumb_track_ratio > 1:
                max_width = int(round(self.width() - 4 * self._thumb_radius))
            else:
                max_width = int(round(self.width() - 4 * self._track_radius))
            # Return early if there is simply no space between thumb positions
            if max_width <= 0:
                self._current_text_width = 0
                self.__display_text = False
                return

            font = self.font()
            font.setPixelSize(max_height)
            metrics = QtGui.QFontMetrics(font)
            text_width = max(
                metrics.horizontalAdvance(f' {text} ') for text in self._state_names if text
            )
            if text_width > max_width:
                text_scale = max_width / text_width
                font.setPixelSize(max(1, int(round(max_height * text_scale))))
            super().setFont(font)

        metrics = QtGui.QFontMetrics(self.font())
        self._current_text_width = max(
            metrics.horizontalAdvance(f' {text} ') for text in self._state_names
        )
        self.__display_text = True

    def _refresh_geometry(self):
        # Calculate new size for track and thumb
        height = self.height()
        if self._thumb_track_ratio > 1:
            self._thumb_radius = height / 2
            self._track_radius = self._thumb_radius / self._thumb_track_ratio
        else:
            self._track_radius = height / 2
            self._thumb_radius = self._track_radius * self._thumb_track_ratio
        self._track_margin = max(0.0, self._thumb_radius - self._track_radius)
        self._thumb_origin = max(self._thumb_radius, self._track_radius)

    def setFont(self, new_font):
        super().setFont(new_font)
        self._refresh_size_hint()
        self._refresh_geometry()
        self._refresh_text_scale()
        self.update()

    @QtCore.Slot()
    def _notify_state_change(self):
        state = self.current_state
        self.sigStateChanged.emit(state if isinstance(state, str) else '')

    @property
    def current_state(self):
        is_checked = self.isChecked()
        return self._state_names[int(is_checked)] if self._state_names else is_checked

    @property
    def _thumb_end(self):
        return self.width() - self._thumb_origin if self.isChecked() else self._thumb_origin

    @property
    def _track_color(self):
        return self._track_colors[int(self.isChecked())]

    @property
    def _thumb_color(self):
        return self._thumb_colors[int(self.isChecked())]

    @property
    def _text_color(self):
        return self._text_colors[int(self.isChecked())]

    @QtCore.Property(float)
    def thumb_position(self):
        return self._thumb_position

    @thumb_position.setter
    def thumb_position(self, value):
        self._thumb_position = value
        self.update()

    def sizeHint(self):
        return self._size_hint

    def minimumSizeHint(self):
        return self._size_hint

    def setChecked(self, checked):
        super().setChecked(checked)
        self._thumb_position = self._thumb_end

    def resizeEvent(self, event):
        self._refresh_geometry()
        self._refresh_text_scale()
        self.thumb_position = self._thumb_end
        event.accept()

    def paintEvent(self, event):
        # Set up painter
        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing, True)
        p.setPen(QtCore.Qt.NoPen)
        track_opacity = self._track_opacity
        if self.isEnabled():
            track_brush = self._track_color
            thumb_brush = self._thumb_color
            text_color = self._text_color
        else:
            palette = self.palette()
            track_opacity *= 0.8
            track_brush = palette.shadow()
            thumb_brush = palette.mid()
            text_color = palette.shadow().color()

        # draw track
        p.setPen(QtCore.Qt.NoPen)
        p.setBrush(track_brush)
        p.setOpacity(track_opacity)
        p.drawRoundedRect(self._track_margin,
                          max(0, self.height() / 2 - self._track_radius),
                          max(0, self.width() - 2 * self._track_margin),
                          2 * self._track_radius,
                          self._track_radius,
                          self._track_radius)

        # draw text if necessary
        if self.__display_text and self._current_text_width > 0:
            p.setPen(text_color)
            p.setOpacity(1.0)
            p.setFont(self.font())
            p.drawText(self._track_margin,
                       self.height() / 2 - self._track_radius,
                       self.width() - 2 * self._track_margin,
                       2 * self._track_radius,
                       QtCore.Qt.AlignCenter,
                       self.current_state)

        # draw thumb
        p.setPen(QtCore.Qt.NoPen)
        p.setBrush(thumb_brush)
        p.setOpacity(1.0)
        p.drawEllipse(self._thumb_position - self._thumb_radius,
                      int(round(self.height()/2 - self._thumb_radius)),
                      2 * self._thumb_radius,
                      2 * self._thumb_radius)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            self._slider_animation.setDuration(200)
            self._slider_animation.setStartValue(self._thumb_position)
            self._slider_animation.setEndValue(self._thumb_end)
            self._slider_animation.start()

    def enterEvent(self, event):
        self.setCursor(QtCore.Qt.PointingHandCursor)
        super().enterEvent(event)



# -*- coding: utf-8 -*-

"""
This file contains models of exponential decay fitting routines for qudi based on the lmfit package.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('ExponentialDecay', 'multiple_exponential_decay')

import warnings
import numpy as np
from scipy.ndimage import filters
from qudi.util.fit_models.model import FitModelBase, estimator


def multiple_exponential_decay(x, amplitudes, decays, stretches):
    """
    Mathematical definition of the sum of multiple stretched exponential decays without any bias.

    WARNING: Iterable parameters "amplitudes", "decays", and "stretches" must have the same length.

    Parameters
    ----------
    x : float
        The independent variable to calculate f(x).
    amplitudes : iterable
        Iterable containing amplitudes for all exponentials.
    decays : iterable
        Iterable containing decay constants for all exponentials.
    stretches : iterable
        Iterable containing stretch constants for all exponentials.

    Returns
    -------
    float or numpy.ndarray
        The result given x for f(x).

    """
    assert len(decays) == len(amplitudes) == len(stretches)
    return sum(amp * np.exp(-(x / decay) ** stretch) for amp, decay, stretch in
               zip(amplitudes, decays, stretches))


class ExponentialDecay(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude', value=1., min=-np.inf, max=np.inf)
        self.set_param_hint('decay', value=1., min=0., max=np.inf)
        self.set_param_hint('stretch', value=1., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, amplitude, decay, stretch):
        return offset + multiple_exponential_decay(x, (amplitude,), (decay,), (stretch,))

    @estimator('Decay')
    def estimate_decay(self, data, x):
        # Smooth very radically the provided data, so that noise fluctuations will not disturb the
        # parameter estimation.
        if len(data) <= 10:
            data_smoothed = data.copy()
        else:
            sigma = max(1, int(round(len(data) / 10)))
            data_smoothed = filters.gaussian_filter1d(data, sigma=sigma)

        # Calculate mean value of first and last 10% of data array. Take the latter as offset.
        mean_len = max(1, len(x) // 10)
        start_mean = np.mean(data_smoothed[mean_len:])
        offset = np.mean(data_smoothed[-mean_len:])

        # subtraction of the offset and normalization of the decay direction
        if start_mean < offset:
            data_smoothed = offset - data_smoothed
        else:
            data_smoothed = data_smoothed - offset

        # Make sure there are no negative values
        smooth_min = min(data_smoothed)
        if smooth_min <= 0:
            data_smoothed -= smooth_min

        # Take all values up to the standard deviation, the remaining values are
        # more disturbing the estimation then helping:
        tmp = np.argwhere(data_smoothed <= np.std(data_smoothed))
        stop_index = tmp[0, 0] if tmp.size > 0 else len(data_smoothed)

        # perform a linear fit on the logarithm of the remaining data
        try:
            poly_coef = np.polyfit(x[:stop_index], np.log(data_smoothed[:stop_index]), deg=1)
            decay = 1 / abs(poly_coef[0])
            amplitude = np.exp(poly_coef[1])
        except:
            warnings.warn('Estimation of decay constant and amplitude failed.')
            decay = abs(data[1] - data[0]) / abs(x[-1] - x[0]) / abs(data[-1] - data[0])
            amplitude = abs(start_mean - offset)

        estimate = self.make_params()
        if start_mean < offset:
            estimate['amplitude'].set(value=-amplitude, max=0)
        else:
            estimate['amplitude'].set(value=amplitude, min=0)
        estimate['offset'].set(value=offset)
        estimate['decay'].set(value=decay, min=2 * min(abs(np.ediff1d(x))))
        estimate['stretch'].set(value=1, vary=False)
        return estimate

    @estimator('Stretched Decay')
    def estimate_stretched_decay(self, data, x):
        estimate = self.estimate_decay(data, x)
        # ToDo: Estimate stretch factor. Currently just a random starting point.
        estimate['stretch'].set(value=2, min=0, max=np.inf, vary=True)
        return estimate

    @estimator('Decay (no offset)')
    def estimate_decay_no_offset(self, data, x):
        estimate = self.estimate_decay(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate

    @estimator('Stretched Decay (no offset)')
    def estimate_stretched_decay_no_offset(self, data, x):
        estimate = self.estimate_stretched_decay(data, x)
        estimate['offset'].set(value=0, min=-np.inf, max=np.inf, vary=False)
        return estimate


class DoubleExponentialDecay(FitModelBase):
    """
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_param_hint('offset', value=0., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_1', value=1., min=-np.inf, max=np.inf)
        self.set_param_hint('amplitude_2', value=1., min=-np.inf, max=np.inf)
        self.set_param_hint('decay_1', value=1., min=0., max=np.inf)
        self.set_param_hint('decay_2', value=1., min=0., max=np.inf)
        self.set_param_hint('stretch_1', value=1., min=0., max=np.inf)
        self.set_param_hint('stretch_2', value=1., min=0., max=np.inf)

    @staticmethod
    def _model_function(x, offset, amplitude_1, amplitude_2, decay_1, decay_2, stretch_1,
                        stretch_2):
        return offset + multiple_exponential_decay(x,
                                                   (amplitude_1, amplitude_2),
                                                   (decay_1, decay_2),
                                                   (stretch_1, stretch_2))



# -*- coding: utf-8 -*-
"""
# FIXME

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

from PySide2 import QtWidgets


class RemoteWidget(QtWidgets.QWidget):
    """

    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        # Create widgets
        local_label = QtWidgets.QLabel('shared modules')
        remote_label = QtWidgets.QLabel('remote modules')
        self.server_label = QtWidgets.QLabel('Server URL')
        self.shared_module_listview = QtWidgets.QListView()
        self.shared_module_listview.setUniformItemSizes(True)
        self.shared_module_listview.setAlternatingRowColors(True)
        self.remote_module_listview = QtWidgets.QListView()
        self.remote_module_listview.setUniformItemSizes(True)
        self.remote_module_listview.setAlternatingRowColors(True)

        # Group widgets in a layout and set as main layout
        layout = QtWidgets.QGridLayout()
        layout.addWidget(self.server_label, 0, 0, 1, 2)
        layout.addWidget(local_label, 1, 0)
        layout.addWidget(self.shared_module_listview, 2, 0)
        layout.addWidget(remote_label, 1, 1)
        layout.addWidget(self.remote_module_listview, 2, 1)
        self.setLayout(layout)



# -*- coding: utf-8 -*-
"""
This file contains Qudi methods for handling real-world values with units.

Copyright (c) 2021-2024, the qudi developers. See the AUTHORS.md file at the top-level directory of
this distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ScalarConstraint', 'DiscreteScalarConstraint']

import warnings
from bisect import bisect_left
from typing import Union, Optional, Tuple, Callable, Any, Iterable

from qudi.util.helpers import is_float, is_integer



class ScalarConstraint:
    """
    Immutable helper object representing the numeric constraints for a scalar number within a
    certain continuous value range.
    Has built-in helper methods to check values against the constraints or clip them to the nearest
    allowed value.

    Attributes
    ----------
    bounds
    minimum
    maximum
    default
    increment
    enforce_int

    Parameters
    ----------
    default : int or float
        Default value to use for this scalar.
    bounds : tuple
        Min and max boundary (inclusive) for the allowed value range.
    increment : int or float, optional
        Default natural step size for value changes. Only used as additional information and not
        for value checking.
    enforce_int : bool, optional
        Flag indicating if this scalar should be enforced to be of integer type
    checker : callable, optional
        Custom checker function to accept a scalar value and raise ValueError or TypeError on fail.
    """
    def __init__(self,
                 default: Union[int, float],
                 bounds: Tuple[Union[int, float], Union[int, float]],
                 increment: Optional[Union[int, float]] = None,
                 enforce_int: Optional[bool] = False,
                 checker: Optional[Callable[[Union[int, float]], bool]] = None
                 ) -> None:
        self._enforce_int = bool(enforce_int)
        self.check_value_type(default)
        for value in bounds:
            self.check_value_type(value)
        if increment is not None:
            self.check_value_type(increment)
        if checker is not None and not callable(checker):
            raise TypeError('checker must be either None or a callable accepting a single scalar '
                            'and returning a valid-flag bool or raising ValueError')
        self._default = default
        self._minimum, self._maximum = sorted(bounds)
        self._increment = increment
        self._checker = checker

        if not self.is_valid(self._default):
            raise ValueError(f'invalid default value ({self._default}) encountered')

    @property
    def bounds(self) -> Tuple[Union[int, float], Union[int, float]]:
        """
        Interval (inclusive) for valid value range.

        Returns
        -------
        int or float
            Minimum allowed value
        int or float
            Maximum allowed value
        """
        return self._minimum, self._maximum

    @property
    def minimum(self) -> Union[int, float]:
        """
        Minimum allowed value.

        Returns
        -------
        int or float
        """
        return self._minimum

    @property
    def maximum(self) -> Union[int, float]:
        """
        Maximum allowed value.

        Returns
        -------
        int or float
        """
        return self._maximum

    @property
    def default(self) -> Union[int, float]:
        """
        Default fallback value.

        Returns
        -------
        int or float
        """
        return self._default

    @property
    def increment(self) -> Union[None, int, float]:
        """
        Natural increment to increase/decrease values. This is often used in a GUI, e.g. QSpinBox.

        Returns
        -------
        int or float
        """
        return self._increment

    @property
    def enforce_int(self) -> bool:
        """
        Flag indicating if this scalar should be enforced to be of integer type

        Returns
        -------
        bool
        """
        return self._enforce_int

    def check(self, value: Union[int, float]) -> None:
        """
        Checks whether the given value is allowed by the constraint by calling various checker
        functions. If a checker function fails it will raise an Exception, indicating what is wrong
        with the checked value.

        Parameters
        ----------
        value : int or float
            value to check

        Raises
        ------
        ValueError
            If the value is incompatible with the constraints
        TypeError
            If the value type is incompatible with the constraints, e.g. in case of
            enforce_int == True
        """
        self.check_value_type(value)
        self.check_value_range(value)
        self.check_custom(value)

    def is_valid(self, value: Union[int, float]) -> bool:
        """
        Checks whether the given value is valid.

        Parameters
        ----------
        value : int or float
            value to check

        Returns
        -------
        bool
        """
        try:
            self.check(value)
        except (ValueError, TypeError):
            return False
        return True

    def clip(self, value: Union[int, float]) -> Union[int, float]:
        """
        Clips the given value to the nearest valid value.

        Parameters
        ----------
        value : int or float
            value to clip

        Returns
        -------
        int or float
        """
        return min(self._maximum, max(self._minimum, value))

    def copy(self) -> object:
        """
        Copies this ScalarConstraint instance.

        Returns
        -------
        ScalarConstraint
            Copy of this instance
        """
        return ScalarConstraint(default=self.default,
                                bounds=self.bounds,
                                increment=self.increment,
                                enforce_int=self.enforce_int,
                                checker=self._checker)

    def check_custom(self, value: Union[int, float]) -> None:
        """
        Checks the given value with the supplied custom checker function.

        Parameters
        ----------
        value : int or float
            value to check

        Raises
        ------
        ValueError
            If custom checker fails to validate.
        TypeError
            If custom checker fails to validate.
        """
        if (self._checker is not None) and (not self._checker(value)):
            raise ValueError(f'Custom checker failed to validate {value}')

    def check_value_range(self, value: Union[int, float]) -> None:
        """
        Checks the given value if it is in bounds.

        Parameters
        ----------
        value : int or float
            value to check

        Raises
        ------
        ValueError
            If value out of bounds.
        """
        if not (self._minimum <= value <= self._maximum):
            raise ValueError(f'Value "{value}" is out of bounds {self.bounds}')

    def check_value_type(self, value: Any) -> None:
        """
        Checks the given value if it is in bounds.

        Parameters
        ----------
        value : int or float
            value to check

        Raises
        ------
        TypeError
            If value is not int or float or if _enforce_int is set if it is not int.
        """
        if self._enforce_int:
            if not is_integer(value):
                raise TypeError(f'values must be int type (received {value})')
        else:
            if not (is_integer(value) or is_float(value)):
                raise TypeError(f'values must be int or float type (received {value})')

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        module = self.__class__.__module__
        return f'{module}.{cls}(' \
               f'default={self.default}, ' \
               f'bounds={self.bounds}, ' \
               f'increment={self.increment}, ' \
               f'enforce_int={self.enforce_int}, ' \
               f'checker={self._checker})'

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict={}):
        new_obj = self.copy()
        memodict[id(self)] = new_obj
        return new_obj

    # Backwards compatibility properties:
    @default.setter
    def default(self, value: Union[int, float]):
        """
        .. deprecated:: 1.3.0
            constraints should be immutable. Pass all values to :py:func:`__init__` instead.
        """
        warnings.warn('ScalarConstraint should be immutable. Pass all values to __init__ instead.',
                      DeprecationWarning,
                      stacklevel=2)
        if not self.is_valid(value):
            raise ValueError(f'invalid default value ({value}) encountered')
        self._default = value

    @property
    def min(self) -> Union[int, float]:
        """
        .. deprecated:: 1.3.0
            ScalarConstraint.min will be removed in the near future. Use ScalarConstraint.minimum
            instead.
        """
        warnings.warn('ScalarConstraint.min will be removed in the near future. '
                      'Use ScalarConstraint.minimum instead.',
                      DeprecationWarning,
                      stacklevel=2)
        return self._minimum

    @min.setter
    def min(self, value: Union[int, float]):
        """
        .. deprecated:: 1.3.0
            constraints should be immutable. Pass all values to :py:func:`__init__` instead.
        """
        warnings.warn('ScalarConstraint should be immutable. Pass all values to __init__ instead.',
                      DeprecationWarning,
                      stacklevel=2)
        self._minimum = value

    @property
    def max(self) -> Union[int, float]:
        """
        .. deprecated:: 1.3.0
            ScalarConstraint.max will be removed in the near future. Use ScalarConstraint.maximum
            instead.
        """
        warnings.warn('ScalarConstraint.max will be removed in the near future. '
                      'Use ScalarConstraint.maximum instead.',
                      DeprecationWarning,
                      stacklevel=2)
        return self._maximum

    @max.setter
    def max(self, value: Union[int, float]):
        """
        .. deprecated:: 1.3.0
            constraints should be immutable. Pass all values to :py:func:`__init__` instead.
        """
        warnings.warn('ScalarConstraint should be immutable. Pass all values to __init__ instead.',
                      DeprecationWarning,
                      stacklevel=2)
        self._maximum = value

    @property
    def step(self) -> Union[None, int, float]:
        """
        .. deprecated:: 1.3.0
            ScalarConstraint.step will be removed in the near future. Use ScalarConstraint.increment
            instead.
        """
        warnings.warn('ScalarConstraint.step will be removed in the near future. '
                      'Use ScalarConstraint.increment instead.',
                      DeprecationWarning,
                      stacklevel=2)
        return self._increment

    @step.setter
    def step(self, value: Union[None, int, float]):
        """
        .. deprecated:: 1.3.0
            constraints should be immutable. Pass all values to :py:func:`__init__` instead.
        """
        warnings.warn('ScalarConstraint should be immutable. Pass all values to __init__ instead.',
                      DeprecationWarning,
                      stacklevel=2)
        self._increment = value


class DiscreteScalarConstraint(ScalarConstraint):
    """
    Specialization of ScalarConstraint for arbitrary discrete sets of allowed scalar values.

    Attributes
    ----------
    allowed_values
    precision
    default
    enforce_int

    Parameters
    ----------
    default : int or float
        Default value to use for this scalar.
    allowed_values : iterable
        Complete collection of allowed scalar values.
    precision : float, optional
        Maximum deviation a checked value is allowed to have from the nearest valid value (default
        is exact matching)
    enforce_int : bool, optional
        Flag indicating if this scalar should be enforced to be of integer type
    checker : callable, optional
        Custom checker function to accept a scalar value and raise ValueError or TypeError on fail.
    """

    def __init__(
        self,
        default: Union[int, float],
        allowed_values: Iterable[Union[int, float]],
        precision: Optional[float] = None,
        enforce_int: Optional[bool] = False,
        checker: Optional[Callable[[Union[int, float]], bool]] = None,
    ) -> None:
        # sort tuple for efficient checking
        if enforce_int:
            allowed_values = (int(val) for val in allowed_values)
        self._allowed_values = tuple(sorted(set(allowed_values)))
        self._precision = None if (precision is None or precision == 0) else abs(precision)
        if len(self._allowed_values) == 0:
            raise ValueError('Must provide at least one allowed value')
        super().__init__(
            default=default,
            bounds=(min(self._allowed_values), max(self._allowed_values)),
            enforce_int=enforce_int,
            checker=checker,
        )

    @property
    def allowed_values(self) -> Tuple[Union[int, float], ...]:
        """
        Discrete collection of values that the constraint allows.

        Returns
        -------
        tuple
        """
        return self._allowed_values

    @property
    def precision(self) -> float:
        """
        Precision with which floating point discrete values are checked for equality.

        Returns
        -------
        float
        """
        return self._precision

    def check(self, value: Union[int, float]) -> None:
        super().check(value)
        self.check_allowed_values(value)

    def check_allowed_values(self, value: Union[int, float]) -> None:
        """
        Method that checks whether the given value is in the set of allowed discrete values.

        Parameters
        ----------
        value : int or float
            value to check

        Raises
        ------
        ValueError
            If value is not in discrete value set.
        """
        closest_allowed = self._find_closest_value(value)
        # Use == operator here since checking for identity can lead to problems with proxy objects
        if closest_allowed != value:
            if self.precision is not None and abs(closest_allowed - value) < self.precision:
                pass
            else:
                raise ValueError(f"Value {value} is not in allowed discrete value set.")

    def clip(self, value: Union[int, float]) -> Union[int, float]:
        return self._find_closest_value(value)

    def copy(self) -> 'DiscreteScalarConstraint':
        """
        Method copies this DiscreteScalarConstraint instance.

        Returns
        -------
        DiscreteScalarConstraint
            Copy of this instance
        """
        return DiscreteScalarConstraint(
            default=self.default,
            allowed_values=self.allowed_values,
            enforce_int=self.enforce_int,
            checker=self._checker,
            precision=self.precision,
        )

    def __repr__(self) -> str:
        """
        Method that gives a readable representation of this constraint.

        Returns
        -------
        str
            Readable representation of the constraint.
        """
        cls = self.__class__.__name__
        module = self.__class__.__module__
        return (
            f"{module}.{cls}("
            f"default={self.default}, "
            f"allowed_values={self.allowed_values}, "
            f"enforce_int={self.enforce_int}, "
            f"checker={self._checker},"
            f"precision={self.precision})"
        )

    def _find_closest_value(self, value: Union[int, float]) -> Union[int, float]:
        """Find allowed value closest to given value"""
        pos = bisect_left(self._allowed_values, value)
        if pos == 0:
            closest = self._allowed_values[pos]
        elif pos == len(self._allowed_values):
            closest = self._allowed_values[-1]
        else:
            before = self._allowed_values[pos - 1]
            after = self._allowed_values[pos]
            if value - before >= after - value:
                closest = after
            else:
                closest = before
        return closest



# -*- coding: utf-8 -*-

"""
This file contains an object representing a qudi configuration.
Qudi configurations are stored in YAML file format.

Copyright (c) 2022, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['Configuration', 'ValidationError', 'ParserError', 'YAMLError', 'DuplicateKeyError']

import copy
from numbers import Number
from PySide2 import QtCore
from typing import Mapping, Optional, Union, Sequence, Set, List, MutableMapping, Any
from collections.abc import MutableMapping as _MutableMapping
from qudi.core.meta import ABCQObjectMeta as _ABCQObjectMeta

from .validator import ValidationError
from .validator import validate_config as _validate_config
from .validator import validate_local_module_config as _validate_local_module_config
from .validator import validate_remote_module_config as _validate_remote_module_config
from .file_handler import ParserError, YAMLError, DuplicateKeyError
from .file_handler import FileHandlerBase as _FileHandlerBase


_OptionType = Union[Sequence, Mapping, Set, Number, str]


class Configuration(_FileHandlerBase,
                    _MutableMapping,
                    QtCore.QObject,
                    metaclass=_ABCQObjectMeta):
    """QObject subclass representing a valid qudi configuration.
    Handles config file loading/dumping as well as writing qudi load config to AppData.
    Performs JSON schema validation upon file loading/dumping and mutation.
    Includes interface methods to add/remove module configurations as well as getting/setting
    various config items.
    """
    sigConfigChanged = QtCore.Signal(object)  # self

    def __init__(self,
                 config: Optional[MutableMapping[str, Any]] = None,
                 parent: Optional[QtCore.QObject] = None
                 ) -> None:
        super().__init__(parent=parent)

        self._file_path = None  # File path for corresponding .cfg file
        self._config = None     # The raw configuration as dict

        # initialize and validate config dict
        self.set_config(config)

    def __repr__(self) -> str:
        return f'Configuration({repr(self._config)})'

    def __str__(self) -> str:
        return f'Configuration({str(self._config)})'

    def __getitem__(self, key: str) -> Any:
        try:
            return copy.deepcopy(self._config['global'][key])
        except KeyError:
            return copy.deepcopy(self._config[key])

    def __setitem__(self, key: str, value: Any) -> None:
        new_config = self.config_map
        if key in new_config:
            new_config[key] = value
        else:
            new_config['global'][key] = value
        self.set_config(new_config)

    def __delitem__(self, key: str) -> None:
        new_config = self.config_map
        try:
            del new_config[key]
        except KeyError:
            del new_config['global'][key]
        self.set_config(new_config)

    def __iter__(self):
        for key, sub_cfg in self.config_map.items():
            if key == 'global':
                yield from sub_cfg
            else:
                yield key

    def __len__(self) -> int:
        return max(0, len(self._config) + len(self._config['global']) - 1)

    @property
    def config_map(self) -> MutableMapping[str, Any]:
        """Deepcopy of the raw config dict."""
        return copy.deepcopy(self._config)

    @property
    def file_path(self) -> Union[None, str]:
        """File path of the associated .cfg file.
        Will be None if no file has been associated, i.e. no load/dump has been performed on this
        config.
        """
        return self._file_path

    def set_config(self, config: Union[None, MutableMapping[str, Any]]) -> None:
        """Validate and reset this Configuration with the given raw config dict.
        """
        new_config = dict() if config is None else copy.deepcopy(config)
        _validate_config(new_config)
        self._config = new_config
        self.sigConfigChanged.emit(self)

    def load(self, file_path: Optional[str] = None, set_default: Optional[bool] = False) -> None:
        """Load a config from file (.cfg), validate it (JSON schema) and reset this Configuration
        instance.

        If no "file_path" argument is given, try to determine the file path by the following
        priority and raise ValueError if no path could be found:
            1. Use "file_path" property if it is not None
            2. Use file path saved in AppData from previous qudi session
            3. Use existing "default.cfg" file in "<UserHome>/qudi/config/"
            4. Use existing "default.cfg" file in "<AppData>/qudi/"
        This instance is only reset on successful JSON schema validation of the loaded config.
        Will set the "file_path" property accordingly on success.

        If optional "set_default" flag is set, write the new file_path to "<AppData>/qudi/load.cfg"
        to be the new default config at the next start of qudi.
        """
        file_path = self._file_path if file_path is None else file_path
        # Try to restore last loaded config file path if possible
        if file_path is None:
            try:
                file_path = self.get_saved_path()
            except FileNotFoundError:
                try:
                    file_path = self.get_default_path()
                except FileNotFoundError:
                    pass
        if file_path is None:
            raise ValueError('No file path defined for configuration to load')

        # Load YAML file from disk.
        config = self._load(file_path)
        # validate config, add missing default values and set as current config.
        old_path = self._file_path
        try:
            self._file_path = file_path
            self.set_config(config)
        except ValidationError:
            self._file_path = old_path
            raise

        # Write current config file path to load.cfg in AppData if requested
        if set_default:
            self.set_default_path(file_path)

    def dump(self, file_path: Optional[str] = None) -> None:
        """Dumps this Configuration instance to file (.cfg) after successful JSON schema
        validation.

        If no "file_path" argument is given, try to use the "file_path" property and raise
        ValueError if no path could be found.
        Will set the "file_path" property accordingly on success.
        """
        file_path = self._file_path if file_path is None else file_path
        if file_path is None:
            raise ValueError('No file path defined for qudi configuration to dump into')
        config = self.config_map
        _validate_config(config)
        self._dump(file_path, config)
        self._file_path = file_path

    def add_local_module(self,
                         base: str,
                         name: str,
                         module_class: str,
                         allow_remote: Optional[bool] = None,
                         connect: Optional[Mapping[str, str]] = None,
                         options: Optional[Mapping[str, _OptionType]] = None) -> None:
        """Mutates the current configuration by validating and adding a new local qudi module
        config with base "gui", "logic" or "hardware" of the form:
            <name>:
                module.Class: <module.Class>
                allow_remote: <allow_remote>
                options:
                    <options_key1>: <options_value1>
                    <options_key2>: <options_value2>
                    ...
                connect:
                    <connect_key1>: <connect_value1>
                    <connect_key2>: <connect_value1>
                    ...

        Raises KeyError if a module with the same name is already configured.
        """
        if self.module_configured(name):
            raise KeyError(f'Module with name "{name}" already configured')
        self.validate_module_base(base)
        module_config = {'module.Class': module_class}
        if allow_remote is not None:
            module_config['allow_remote'] = allow_remote
        if connect is not None:
            module_config['connect'] = copy.copy(connect)
        if options is not None:
            module_config['options'] = copy.deepcopy(options)
        _validate_local_module_config(module_config)
        new_config = self.config_map
        new_config[base][name] = module_config
        self.set_config(new_config)

    def add_remote_module(self,
                          base: str,
                          name: str,
                          native_module_name: str,
                          address: str,
                          port: int,
                          certfile: Optional[str] = None,
                          keyfile: Optional[str] = None) -> None:
        """Mutates the current configuration by validating and adding a new remote qudi module
        config with base "gui", "logic" or "hardware" of the form:
            <name>:
                native_module_name: <native_module_name>
                address: <address>
                port: <port>
                certfile: <certfile>
                keyfile: <keyfile>

        Raises KeyError if a module with the same name is already configured.
        """
        if self.module_configured(name):
            raise KeyError(f'Module with name "{name}" already configured')
        self.validate_module_base(base)
        module_config = {'native_module_name': native_module_name,
                         'address'           : address,
                         'port'              : port}
        if certfile is not None:
            module_config['certfile'] = certfile
        if keyfile is not None:
            module_config['keyfile'] = keyfile
        _validate_remote_module_config(module_config)
        new_config = self.config_map
        new_config[base][name] = module_config
        self.set_config(new_config)

    def rename_module(self, old_name: str, new_name: str) -> None:
        """Mutates the current configuration by validating and renaming an already configured
        module.

        Raises KeyError if a module with <new_name> is already configured or if <old_name> can not
        be found in the current config.
        """
        if old_name == new_name:
            return
        if not self.module_configured(old_name):
            raise KeyError(f'No module with name "{old_name}" configured')
        if self.module_configured(new_name):
            raise KeyError(f'Module with name "{new_name}" already configured')

        new_config = self.config_map
        for base in ['gui', 'logic', 'hardware']:
            try:
                module_config = new_config[base].pop(old_name)
            except KeyError:
                continue
            else:
                new_config[base][new_name] = module_config
                self.set_config(new_config)
                return

    def remove_module(self, name: str) -> None:
        """Mutates the current configuration by deleting a configured module.

        Raises KeyError if no module is configured by given <name>.
        """
        new_config = self.config_map
        for base in ['gui', 'logic', 'hardware']:
            try:
                del new_config[base][name]
            except KeyError:
                continue
            else:
                self.set_config(new_config)
                return
        raise KeyError(f'No module with name "{name}" configured')

    def module_configured(self, name: str) -> bool:
        """Checks if a module with given name is present in current configuration."""
        return name in self._config['gui'] or name in self._config['logic'] or name in self._config[
            'hardware']

    def module_config(self, name: str) -> MutableMapping[str, Any]:
        """Returns module configuration for given module <name>.

        Raises KeyError if no module is configured by given <name>.
        """
        for base in ['gui', 'logic', 'hardware']:
            try:
                return copy.deepcopy(self._config[base][name])
            except KeyError:
                pass
        raise KeyError(f'No module with name "{name}" configured')

    def is_remote_module(self, name: str) -> bool:
        """Checks whether a configured module is a remote module and returns answer flag.

        Raises KeyError if no module is configured by given <name>.
        """
        return 'native_module_name' in self.get_module_config(name)

    def is_local_module(self, name):
        """Checks whether a configured module is a local module and returns answer flag.

        Raises KeyError if no module is configured by given <name>.
        """
        return 'module.Class' in self.get_module_config(name)

    @property
    def module_names(self) -> List[str]:
        """List of the currently configured module names."""
        return [*self._config['gui'], *self._config['logic'], *self._config['hardware']]

    @staticmethod
    def validate_module_base(base: str) -> None:
        """Raises ValueError if the given string is no valid qudi module base."""
        if base not in ['gui', 'logic', 'hardware']:
            raise ValueError('qudi module base must be one of ["gui", "logic", "hardware"]')



# -*- coding: utf-8 -*-
"""
Definition of various metaclasses

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('ABCQObjectMeta', 'ModuleMeta', 'QObjectMeta', 'QudiObjectMeta')

from abc import ABCMeta
from PySide2.QtCore import QObject
from qudi.core.statusvariable import StatusVar
from qudi.core.connector import Connector
from qudi.core.configoption import ConfigOption


QObjectMeta = type(QObject)


class ABCQObjectMeta(ABCMeta, QObjectMeta):
    """Metaclass for abstract QObject subclasses.
    """

    def __new__(mcs, name, bases, attributes):
        cls = super(ABCQObjectMeta, mcs).__new__(mcs, name, bases, attributes)
        # Compute set of abstract method names
        abstracts = {
            attr_name for attr_name, attr in attributes.items() if \
            getattr(attr, '__isabstractmethod__', False)
        }
        for base in bases:
            for attr_name in getattr(base, '__abstractmethods__', set()):
                attr = getattr(cls, attr_name, None)
                if getattr(attr, '__isabstractmethod__', False):
                    abstracts.add(attr_name)
        cls.__abstractmethods__ = frozenset(abstracts)
        return cls


class QudiObjectMeta(ABCQObjectMeta):
    """General purpose metaclass for abstract QObject subclasses that include qudi meta objects
    (Connector, StatusVar, ConfigOption).
    Collects all meta objects in new "_meta" class variable for easier access.
    """
    def __new__(mcs, name, bases, attributes):
        cls = super().__new__(mcs, name, bases, attributes)

        meta = dict()

        # Collect qudi module meta attributes (Connector, StatusVar, ConfigOption) and put them
        # in the class variable dict "_meta" for easy bookkeeping and access.
        connectors = dict()
        status_vars = dict()
        config_opt = dict()
        for attr_name in dir(cls):
            attr = getattr(cls, attr_name, None)
            if isinstance(attr, Connector):
                connectors[attr_name] = attr
            elif isinstance(attr, StatusVar):
                status_vars[attr_name] = attr
            elif isinstance(attr, ConfigOption):
                config_opt[attr_name] = attr
        meta.update({'connectors'      : connectors,
                     'status_variables': status_vars,
                     'config_options'  : config_opt})
        setattr(cls, '_meta', meta)
        return cls


class ModuleMeta(QudiObjectMeta):
    """Metaclass for all qudi modules (GUI, logic and hardware).
    """

    def __new__(mcs, name, bases, attributes):
        cls = super().__new__(mcs, name, bases, attributes)

        # Determine module base key and add to _meta dict
        if getattr(cls, '_meta', None):
            for base in cls.mro():
                if base.__name__ == 'GuiBase':
                    cls._meta['base'] = 'gui'
                    break
                elif base.__name__ == 'LogicBase':
                    cls._meta['base'] = 'logic'
                    break
                elif base.__name__ == 'Base':
                    cls._meta['base'] = 'hardware'
                    break
        return cls
